import argparse
from pathlib import Path

import numpy as np
import cv2 as cv
import glob
import random
import tqdm


def main(directory):
    """https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html"""
    pattern_size = (11, 8)
    max_num_images = 70

    # termination criteria
    criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
    objp = np.zeros((pattern_size[1]*pattern_size[0],3), np.float32)
    objp[:,:2] = np.mgrid[0:pattern_size[0], 0:pattern_size[1]].T.reshape(-1,2)
    # Arrays to store object points and image points from all the images.
    objpoints = [] # 3d point in real world space
    imgpoints = [] # 2d points in image plane.
    images = random.sample([
        i for i in glob.glob(str(directory / '*.png'))
        if Path(i).name.startswith('color_')
    ], max_num_images)

    for fname in tqdm.tqdm(images):
        img = cv.imread(fname)
        gray = cv.cvtColor(img, cv.COLOR_BGR2GRAY)
        h, w = gray.shape
        # Find the chess board corners
        ret, corners = cv.findChessboardCorners(gray, pattern_size, None)

        # If found, add object points, image points (after refining them)
        if ret == True:
            objpoints.append(objp)
            corners2 = cv.cornerSubPix(gray,corners, (11,11), (-1,-1), criteria)
            imgpoints.append(corners)

            # Draw and display the corners
            cv.drawChessboardCorners(img, pattern_size, corners2, ret)
            cv.imshow('img', img)
            cv.waitKey(50)
    cv.destroyAllWindows()

    ret, mtx, dist, rvecs, tvecs = cv.calibrateCamera(
        objpoints, imgpoints, gray.shape[::-1], None, None
    )

    print(mtx)
    print(dist)

    np.save('color_camK.npy', mtx)
    np.save('color_cam_distortion_coeffs.npy', dist)

    # print(ret)
    # print(rvecs, tvecs)

    # newcameramtx, roi = cv.getOptimalNewCameraMatrix(
    #     mtx, dist, (w,h), 1, (w,h)
    # )
    # print(newcameramtx, roi)

    ## Re-projection Error
    mean_error = 0
    for i in range(len(objpoints)):
        imgpoints2, _ = cv.projectPoints(objpoints[i], rvecs[i], tvecs[i], mtx, dist)
        error = cv.norm(imgpoints[i], imgpoints2, cv.NORM_L2)/len(imgpoints2)
        mean_error += error
    print( "total error: {}".format(mean_error/len(objpoints)) )

    ## Undistortion
    dst = cv.undistort(img, mtx, dist)
    concatenated_img = np.concatenate((img, dst), axis=1)
    cv.imwrite('/tmp/calibration.png', concatenated_img)
    cv.namedWindow('img', cv.WINDOW_NORMAL)
    cv.imshow('img', concatenated_img)
    cv.waitKey()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'directory', type=Path,
    )
    args = parser.parse_args()
    main(**args.__dict__)
