import numpy as np
import open3d as o3d
import matplotlib.pyplot as plt
from tqdm import tqdm
import os
import tempfile
import cv2

from .utils.logging import get_main_logger
from .utils.constant import NUM_IMAGE
from .utils.registration import full_registration_and_combine, merge_pcds
from .utils.vis import get_random_color, generate_scene_pointcloud

logger = get_main_logger()

class TOC_Dataset():
    """Toc Dataset toolkit class

    Args:
        root(str): root path for the dataset.
    """
    def __init__(self, root):
        self.root = root
        self.scene_name_list = self.load_scene_name_list()
        self.object_name_list, self.object_id_dict = self.load_object_list()

        logger.info('obj name list:{}\nobj id dict:{}\nscene name list:{}'.format(
            self.object_name_list,
            self.object_id_dict,
            self.scene_name_list
            )
        )

    def load_object_list(self, save = False):
        if os.path.exists(os.path.join(self.root, 'object_name_list.txt')):
            f = open(os.path.join(self.root, 'object_name_list.txt'))
            object_name_list = f.readlines()
            f.close()
            object_id_dict = dict()
            for i in range(len(object_name_list)):
                object_name_list[i] = object_name_list[i].strip()
                object_id_dict[object_name_list[i]] = i + 1
            return object_name_list, object_id_dict
        else:
            logger.warning('No existed object list, generate object list from scenes')
            object_name_set = set()
            for scene_id in range(len(self.scene_name_list)):
                scene_object_name_list = self.load_scene_object_list(scene_id)
                for object_name in scene_object_name_list:
                    object_name_set.add(object_name)
            object_name_list = list(object_name_set)
            object_id_dict = dict()
            for i in range(len(object_name_list)):
                object_id_dict[object_name_list[i]] = i + 1
            if save:
                logger.warning('Save object list to dataset folder')
                f = open(os.path.join(self.root, 'object_name_list.txt'), 'w')
                for object_name in object_name_list:
                    f.write(object_name+'\n')
                f.close()
            return object_name_list, object_id_dict

    def load_scene_name_list(self):
        scene_list_file_path = os.path.join(self.root, 'scene_name_list.txt')
        if not os.path.exists(scene_list_file_path):
            logger.warning("No existed scene list, generate scene list from scenes")
            scene_name_list = os.listdir(os.path.join(self.root, 'scenes'))
        else:
            f = open(scene_list_file_path)
            scene_name_list = f.readlines()
            f.close()
            for i in range(len(scene_name_list)):
                scene_name_list[i] = scene_name_list[i].strip()
        return scene_name_list

    def load_scene_name(self, scene_id):
        return self.scene_name_list[scene_id]

    def load_camera_param(self, scene_id):
        camera_intrin_matrix = np.load(
            os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'color_camK.npy'
            )
        )
        param = o3d.camera.PinholeCameraParameters()
        param.extrinsic = np.eye(4,dtype=np.float64)
        param.intrinsic.set_intrinsics(
            1280,
            720,
            camera_intrin_matrix[0][0],
            camera_intrin_matrix[1][1],
            camera_intrin_matrix[0][2],
            camera_intrin_matrix[1][2],
        )
        return param

    def load_real_camera_intrinsic(self, scene_id):
        return np.load(
            os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'color_camK.npy'
            )
        )

    def load_raw_image(self, scene_id, image_id, order = 'RGB'):
        raw_bgr = cv2.imread(
            os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'rgb_undistort',
                # '%04d.png' % image_id
                'color_%d.png' % image_id
            )
        )
        if order == 'BGR':
            return raw_bgr
        elif order == 'RGB':
            return cv2.cvtColor(raw_bgr, cv2.COLOR_BGR2RGB)
        else:
            raise ValueError('Unknown order {}, only RGB and BGR are allowed.'.format(order))

    def load_depth_image(self, scene_id, image_id):
        depth = cv2.imread(
            os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'depth_undistort',
                '%04d.png' % image_id
            ),
            cv2.IMREAD_UNCHANGED
        )
        return depth

    def load_camera_pose(self, scene_id, image_id):
        return np.load(os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'camera_pose',
                '%04d.npy' % image_id
            )
        )

    def load_scene_object_list(self, scene_id):
        f = open(
            os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'object_list.txt'
            )
        )
        scene_object_name_list = f.readlines()
        f.close()
        for i in range(len(scene_object_name_list)):
            scene_object_name_list[i] = scene_object_name_list[i].strip()
        return scene_object_name_list

    def load_object_mesh(self, model_name, with_color = True):
        if with_color:
            o3d_mesh = o3d.io.read_triangle_mesh(
                os.path.join(
                    self.root,
                    'rgb_pcd',
                    '{}.ply'.format(model_name)
                )
            )
        else:
            o3d_mesh = o3d.io.read_triangle_mesh(
                os.path.join(
                    self.root,
                    'models',
                    model_name,
                    'visual_meshes',
                    'visual.ply'
                )
            )

        return o3d_mesh

    def load_object_pose_dict(self, scene_id):
        object_pose_dict = dict()
        for object_name in self.load_scene_object_list(scene_id):
            pose_file_path = os.path.join(
                    self.root,
                    'scenes',
                    self.load_scene_name(scene_id),
                    'obj_poses',
                    object_name+'.npy'
                )
            if os.path.exists(pose_file_path):
                object_pose_dict[object_name] = np.load(pose_file_path)
            else:
                object_pose_dict[object_name] = None
        return object_pose_dict

    def load_per_frame_object_pose(self, scene_id, image_id):
        object_pose = np.load(
            os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'obj_poses_per_frame',
                '%04d.npy' % image_id
            ),
            allow_pickle = True
        ).item()
        return object_pose

    def generate_seg_mask(self, scene_id, image_id, save = False):
        object_pose_dict = self.load_object_pose_dict(scene_id)
        camera_param = self.load_camera_param(scene_id)
        scene_object_list = self.load_scene_object_list(scene_id)
        camera_pose = self.load_camera_pose(scene_id, image_id)
        geometry_list = []
        for scene_object_name in scene_object_list:
            if object_pose_dict[scene_object_name] is not None:
                o3d_mesh = self.load_object_mesh(scene_object_name)
                sixd_pose = object_pose_dict[scene_object_name]
                o3d_mesh.transform(sixd_pose)
                o3d_mesh.transform(np.linalg.inv(camera_pose))
                obj_mask_id = (self.object_id_dict[scene_object_name])/ 255
                o3d_mesh.paint_uniform_color([obj_mask_id, obj_mask_id, obj_mask_id])
                geometry_list.append(o3d_mesh)
        vis = o3d.visualization.Visualizer()
        vis.create_window(width = 1280, height = 720)
        for g in geometry_list:
            vis.add_geometry(g)

        ctr = vis.get_view_control()
        ctr.convert_from_pinhole_camera_parameters(camera_param)
        vis.poll_events()
        tmp_file = tempfile.mkstemp('.png')
        vis.capture_screen_image(tmp_file[1], do_render=False)
        seg_mask = cv2.imread(tmp_file[1])[:,:,0]
        os.system('rm {}'.format(tmp_file[1]))
        valid_mask = np.zeros((720, 1280), dtype = bool)
        for scene_object_name in scene_object_list:
            obj_id = self.object_id_dict[scene_object_name]
            valid_mask = valid_mask | (seg_mask == (obj_id))
        seg_mask[~valid_mask] = 0
        vis.destroy_window()
        vis.close()
        if save:
            seg_mask_dir = os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'seg_masks',
            )
            if not os.path.exists(seg_mask_dir):
                os.makedirs(seg_mask_dir)
            seg_mask_path = os.path.join(seg_mask_dir, '%04d.npy' % image_id)
            logger.warning('Save segment to {}'.format(seg_mask_path))
            np.save(seg_mask_path, seg_mask)
        return seg_mask

    def load_seg_mask(self, scene_id, image_id):
        seg_mask_path = os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'seg_masks',
                '%04d.npy' % image_id
            )
        return np.load(seg_mask_path)

    def load_point_cloud(self, scene_id, image_id):
        pcd = generate_scene_pointcloud(
            depth_path = os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'depth_undistort',
                '%04d.png' % image_id
            ),
            rgb_path = os.path.join(
                self.root,
                'scenes',
                self.load_scene_name(scene_id),
                'rgb_undistort',
                '%04d.png' % image_id
            ),
            intrinsics = self.load_real_camera_intrinsic(scene_id),
            depth_scale = 1000.0
        )
        return pcd

    def load_scene_point_cloud(self, scene_id):
        pcds = []
        # Only half of the scenes are used to calculate the full scene
        for image_id in range(0, NUM_IMAGE, 4):
            income_pcd = self.load_point_cloud(scene_id = scene_id, image_id = image_id)
            camera_pose = self.load_camera_pose(scene_id = scene_id, image_id = image_id)
            income_pcd.transform(camera_pose)
            pcds.append(income_pcd)
        full_pcd = full_registration_and_combine(pcds)
        return full_pcd

    def vis_6dpose(self, scene_id, image_id = None, dimension = 3, show = True, per_frame = False):
        """Visualize 6d pose annotation in a scene or in an image

        Args:
            scene_id(int): the id of the scene.
            image_id(int or None): the id of the image, None for the whole scene.
            dimension(int): 2 for 2d visualization and 3 for 3d visualization.
            show(bool): whether to show the result using open3d.
            per_frame(bool): whether to use perframe 6d pose annotation.

        Returns:
            o3d.geometry.PointCloud or np.array: If dimension==2, returns point cloud; elif
            dimension == 3, returns BGR image.
        """
        if not per_frame:
            raise ValueError('Currently, only per frame pose is availabel')
        if dimension == 2:
            if image_id is not None:
                bgr = self.load_raw_image(scene_id, image_id, order = 'BGR')
                mask = self.load_seg_mask(scene_id, image_id)
                scene_object_list = self.load_scene_object_list(scene_id)
                for scene_object_name in scene_object_list:

                    object_id = self.object_id_dict[scene_object_name]
                    instance_mask = mask == object_id

                    bgr[instance_mask] = bgr[instance_mask] // 2 + get_random_color() // 2
                if show:
                    cv2.imshow('{}-{} vis'.format(scene_id, image_id), bgr)
                    cv2.waitKey(0)
                    cv2.destroyAllWindows()
                return bgr
            else:
                raise ValueError('The image id must be given for 2d visualization')
        elif dimension == 3:
            geometry_list = []
            if image_id is not None:
                image_pcd = self.load_point_cloud(scene_id, image_id)
                camera_pose = self.load_camera_pose(scene_id, image_id)
                if not per_frame:
                    image_pcd.transform(camera_pose)
                geometry_list.append(image_pcd)
            else:
                scene_pcd = self.load_scene_point_cloud(scene_id)
                geometry_list.append(scene_pcd)
            if per_frame:
                object_pose_dict= self.load_per_frame_object_pose(scene_id, image_id)
            else:
                object_pose_dict = self.load_object_pose_dict(scene_id)
            scene_object_list = self.load_scene_object_list(scene_id)
            for scene_object_name in scene_object_list:
                if (scene_object_name in object_pose_dict.keys()) and (object_pose_dict[scene_object_name] is not None):
                    o3d_mesh = self.load_object_mesh(scene_object_name)
                    sixd_pose = object_pose_dict[scene_object_name]
                    o3d_mesh.transform(sixd_pose)
                    geometry_list.append(o3d_mesh.sample_points_uniformly(10000))
            full_pcd = merge_pcds(geometry_list)
            if show:
                o3d.visualization.draw_geometries([full_pcd])
            return full_pcd
        else:
            raise ValueError('Dimension must be 2 or 3')
