import open3d as o3d
import numpy as np
from .logging import get_main_logger

logger = get_main_logger()

default_voxel_size = 0.005

def pairwise_registration(source, target, voxel_size = default_voxel_size):
    logger.debug("Apply point-to-plane ICP")
    max_correspondence_distance_coarse = voxel_size * 15
    max_correspondence_distance_fine = voxel_size * 1.5
    icp_coarse = o3d.pipelines.registration.registration_icp(
        source, target, max_correspondence_distance_coarse, np.identity(4),
        o3d.pipelines.registration.TransformationEstimationPointToPlane()
    )
    icp_fine = o3d.pipelines.registration.registration_icp(
        source, target, max_correspondence_distance_fine,
        icp_coarse.transformation,
        o3d.pipelines.registration.TransformationEstimationPointToPlane()
    )
    transformation_icp = icp_fine.transformation
    information_icp = o3d.pipelines.registration.get_information_matrix_from_point_clouds(
        source, target, max_correspondence_distance_fine,
        icp_fine.transformation
    )
    return transformation_icp, information_icp

def full_registration(pcds, voxel_size = default_voxel_size):
    pose_graph = o3d.pipelines.registration.PoseGraph()
    odometry = np.identity(4)
    pose_graph.nodes.append(o3d.pipelines.registration.PoseGraphNode(odometry))
    n_pcds = len(pcds)
    for source_id in range(n_pcds):
        for target_id in range(source_id + 1, n_pcds):
            logger.debug('full registration: source:{}/{} target:{}/{}'.format(
                source_id + 1,
                n_pcds,
                target_id - source_id,
                n_pcds - source_id - 1
            ))
            transformation_icp, information_icp = pairwise_registration(
                pcds[source_id],
                pcds[target_id],
                voxel_size = voxel_size
            )
            logger.debug("Build o3d.pipelines.registration.PoseGraph")
            if target_id == source_id + 1:  # odometry case
                odometry = np.dot(transformation_icp, odometry)
                pose_graph.nodes.append(
                    o3d.pipelines.registration.PoseGraphNode(
                        np.linalg.inv(odometry)
                    )
                )
                pose_graph.edges.append(
                    o3d.pipelines.registration.PoseGraphEdge(
                        source_id,
                        target_id,
                        transformation_icp,
                        information_icp,
                        uncertain=False
                    )
                )
            else:
                pose_graph.edges.append(
                    o3d.pipelines.registration.PoseGraphEdge(
                        source_id,
                        target_id,
                        transformation_icp,
                        information_icp,
                        uncertain=True
                    )
                )
    return pose_graph

def optimize(pose_graph, voxel_size = default_voxel_size):
    logger.debug("optimize")
    max_correspondence_distance_coarse = voxel_size * 15
    max_correspondence_distance_fine = voxel_size * 1.5
    option = o3d.pipelines.registration.GlobalOptimizationOption(
        max_correspondence_distance=max_correspondence_distance_fine,
        edge_prune_threshold=0.9,
        reference_node=0
    )
    o3d.pipelines.registration.global_optimization(
        pose_graph,
        o3d.pipelines.registration.GlobalOptimizationLevenbergMarquardt(),
        o3d.pipelines.registration.GlobalOptimizationConvergenceCriteria(),
        option
    )
    return pose_graph

def full_registration_and_combine(pcds, voxel_size = default_voxel_size):
    logger.debug('full scene pcd: begin preprocess')
    logger.warning('Currently, point clouds are directly combined without registration, use with care!')
    # for i in range(len(pcds)):
    #     pcds[i] = pcds[i].voxel_down_sample(voxel_size)
    #     if not pcds[i].has_normals():
    #         pcds[i].estimate_normals(o3d.geometry.KDTreeSearchParamKNN(30))
    # logger.debug('full scene pcd: begin full registration')
    # pose_graph = full_registration(pcds, voxel_size = voxel_size)
    # logger.debug('full scene pcd: begin optimize')
    # pose_graph = optimize(pose_graph, voxel_size = voxel_size)
    # for point_id in range(len(pcds)):
    #     pcds[point_id].transform(pose_graph.nodes[point_id].pose)
    out_pcd = merge_pcds(pcds).voxel_down_sample(voxel_size)
    out_pcd = out_pcd.remove_statistical_outlier(3000, 0.5)[0]
    return out_pcd

def merge_pcds(pcds):
    points = np.zeros(shape = (0, 3), dtype = np.float64)
    colors = np.zeros(shape = (0, 3), dtype = np.float64)
    for pcd in pcds:
        points = np.vstack((points, np.asarray(pcd.points)))
        colors = np.vstack((colors, np.asarray(pcd.colors)))
    out_pcd = o3d.geometry.PointCloud()
    out_pcd.points = o3d.utility.Vector3dVector(points)
    out_pcd.colors = o3d.utility.Vector3dVector(colors)
    return out_pcd