#!/usr/bin/env python
import sys
import os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'dataset_toolkit'))

import argparse
from pathlib import Path
import tqdm

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level
from renderer import resize_camera_parameters

import cv2
import numpy as np
import open3d as o3d
from pyquaternion import Quaternion


def main(dataset_root, scene_name, resize_factor, num_images):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    scene_dir = dataset_root / 'scenes' / scene_name

    camera_poses = get_orb_poses(scene_dir, num_images)
    camera_param = dataset.load_camera_param(scene_id)
    camera_param = resize_camera_parameters(camera_param, resize_factor)
    print(
        f'width: {camera_param.intrinsic.width}, '
        f'height: {camera_param.intrinsic.height}, '
        f'f: {camera_param.intrinsic.get_focal_length()}, '
        f'c: {camera_param.intrinsic.get_principal_point()}, '
    )

    volume = volume_reconstruction(
        scene_dir, resize_factor, camera_poses, camera_param,
        voxel_length=4.0 / 512.0, depth_trunc=4.0
    )
    mesh = volume.extract_triangle_mesh()
    mesh.compute_vertex_normals()
    o3d.io.write_triangle_mesh(
        str(scene_dir / 'fastfusion_reconstruction.ply'), mesh
    )

    if (scene_dir / 'mask_rcnn_masks').is_dir():
        volume = volume_reconstruction(
            scene_dir, resize_factor, camera_poses, camera_param,
            voxel_length=2.5 / 512.0, depth_trunc=2.5, mask_depth=True
        )
        mesh = volume.extract_triangle_mesh()
        mesh.compute_vertex_normals()
        o3d.io.write_triangle_mesh(
            str(scene_dir / 'object_reconstruction.ply'), mesh
        )


def volume_reconstruction(
        scene_dir, resize_factor, camera_poses, camera_params,
        voxel_length, depth_trunc, mask_depth=False
):
    volume = o3d.pipelines.integration.ScalableTSDFVolume(
        voxel_length=voxel_length,
        sdf_trunc=0.04,
        color_type=o3d.pipelines.integration.TSDFVolumeColorType.RGB8
    )

    for camera_idx, camera_pose in tqdm.tqdm(camera_poses.items()):
        color = o3d.io.read_image(
            str(scene_dir / 'rgb_undistort' / f'color_{camera_idx}.png'),
        )

        depth = cv2.imread(
            str(scene_dir / 'distorted_images' / f'depth_{camera_idx}.png'),
            cv2.IMREAD_UNCHANGED
        )

        if resize_factor != 1:
            new_size = [int(i/resize_factor) for i in depth.shape[::-1]]
            depth = cv2.resize(depth, new_size)

        if mask_depth:
            mask_path = scene_dir / 'mask_rcnn_masks' / f'color_{camera_idx}.png'
            if not mask_path.is_file():
                continue

            mask = cv2.imread(str(mask_path), cv2.IMREAD_UNCHANGED)
            depth = depth * (mask > 0)

        depth = o3d.geometry.Image((depth).astype(np.float32))

        rgbd = o3d.geometry.RGBDImage.create_from_color_and_depth(
            color, depth, convert_rgb_to_intensity=False,
            depth_scale=1000.0, depth_trunc=depth_trunc
        )

        volume.integrate(
            rgbd, camera_params.intrinsic, np.linalg.inv(camera_pose)
        )

    return volume


def get_orb_poses(scene_dir, num_images):
    # orb_poses_path = scene_dir / 'KeyFrameTrajectory.txt'
    orb_poses_path = scene_dir / 'CameraTrajectory.txt'
    orb_poses = np.loadtxt(orb_poses_path)

    total_images = len(orb_poses)
    if (num_images <= 0) or (total_images <= num_images):
        step_size = 1
    else:
        step_size = int(np.floor(float(total_images) / num_images))

    print(f'total_images: {total_images}, num_images: {num_images}')
    print(f'step_size: {step_size}, num_images: {len(orb_poses[::step_size])}')

    T_world_cams = {}
    for orb_pose in orb_poses[::step_size]:
        cam_idx = int(orb_pose[0])
        T_world_cam = orb_pose_to_mat(orb_pose)
        T_world_cams[cam_idx] = T_world_cam

    return T_world_cams


def orb_pose_to_mat(pose):
    x, y, z, qx, qy, qz, qw = pose[1:].tolist()
    T_world_cam = Quaternion(qw, qx, qy, qz).transformation_matrix
    T_world_cam[:3, -1] = [x, y, z]
    return T_world_cam


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=1)
    parser.add_argument('--num-images', type=int, default=-1)
    args = parser.parse_args()
    main(**args.__dict__)
