#!/usr/bin/env python
import sys
sys.path.append('dataset_toolkit')

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

from align_utils import pose_align, correct_rotation

from pose_estimation import viz_objects, get_depth_point_cloud
from pose_estimation import icp_refine_object_poses, get_kinematic_cam_poses

from constants import get_workspace_bounding_box

import argparse
from pathlib import Path
import numpy as np
import copy

import open3d as o3d
import matplotlib.pyplot as plt


RADIUS_FILTER_NB_POINTS = 20
RADIUS_FILTER_RADIUS = 0.03

DEPTH_CLOUD_RESOLUTION = 0.001

CLUSTER_DENSITY = 0.005
CLUSTER_MIN_POINTS = 150

# ICP alignment threshold
ICP_ALIGNMENT_THRESHOLD = 0.005


def main(dataset_root, scene_name):
    set_log_level('WARNING')
    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    # scene_name = scene_name_list[scene_id]

    print('scene', scene_id, scene_name)

    kinematic_cam_poses = get_kinematic_cam_poses(dataset, scene_name)

    est_cam_poses = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'camera_poses.npy'),
        allow_pickle=True
    )[()]
    est_obj_poses = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'object_poses.npy'),
        allow_pickle=True
    )[()]


    # aligned_cam_poses, alignment_transform = pose_align(
    #     kinematic_cam_poses, est_cam_poses
    # )
    # aligned_obj_poses = {
    #     key: alignment_transform.dot(pose)
    #     for key, pose in est_obj_poses.items()
    # }

    # already aligned
    aligned_cam_poses = est_cam_poses
    aligned_obj_poses = est_obj_poses
    alignment_transform = np.load(
        str(Path(dataset.root) / 'scenes' / scene_name / 'alignment_transform.npy'),
    )

    scene_object_list = dataset.load_scene_object_list(scene_id)
    unknown_objects = [i for i in scene_object_list if i not in aligned_obj_poses]
    # unknown_objects = ['suger_1']
    print('unknown_objects', unknown_objects)

    for unknown_obj in unknown_objects:
        if unknown_obj in aligned_obj_poses:
            del aligned_obj_poses[unknown_obj]

    ## using MVS point cloud
    # dense_colmap_output = dataset_root / 'scenes' / scene_name / 'dense_model2'
    # point_cloud \
    #     = o3d.io.read_point_cloud(str(dense_colmap_output / 'fused.ply')) \
    #             .transform(alignment_transform)

    ## using depth point cloud
    depth_cloud_file = dataset_root / 'scenes' / scene_name / 'depth_cloud.ply'

    # point_cloud = get_depth_point_cloud(dataset, scene_id, aligned_cam_poses)
    # # lots of redundant points, downsaple for efficiency
    # point_cloud = point_cloud.voxel_down_sample(voxel_size=DEPTH_CLOUD_RESOLUTION)
    # o3d.io.write_point_cloud(str(depth_cloud_file), point_cloud)

    point_cloud = o3d.io.read_point_cloud(str(depth_cloud_file))

    workspace_bb = get_workspace_bounding_box()
    point_cloud = point_cloud.crop(workspace_bb)
    # point_cloud, inlier_ind = point_cloud.remove_radius_outlier(
    #     nb_points=RADIUS_FILTER_NB_POINTS, radius=RADIUS_FILTER_RADIUS
    # )
    point_cloud = remove_objects_from_point_cloud(
        dataset, point_cloud, aligned_obj_poses
    )
    clusters, cluster_colored_pcd = cluster_point_cloud(point_cloud)
    # point_cloud = cluster_colored_pcd
    if len(clusters):
        point_cloud = clusters[0]

    unknown_obj_poses, pose_debugging = localize_objects(
        dataset, scene_id, unknown_objects, clusters
    )

    aligned_obj_poses = {**aligned_obj_poses, **unknown_obj_poses}

    viz_objects(
        dataset, scene_id, aligned_obj_poses,
        aligned_cam_poses, None, point_cloud
    )

    out_scene_dir=Path(dataset.root)/ 'scenes'/ scene_name
    np.save(str(out_scene_dir/'aligned_camera_poses.npy'), aligned_cam_poses)
    np.save(str(out_scene_dir/'aligned_object_poses.npy'), aligned_obj_poses)

    ## debugging
    # aligned_obj_poses[unknown_objects[-1]] \
    #     = pose_debugging['pose_candidates'][3]

    # geometry_list = get_geometry_list(
    #     dataset, scene_id, aligned_obj_poses, aligned_cam_poses, point_cloud
    # )

    # geometry_list.extend([
    #     o3d.geometry.LineSet.create_from_oriented_bounding_box(
    #         pose_debugging['object_obbs'][3]
    #     ),
    #     o3d.geometry.LineSet.create_from_oriented_bounding_box(
    #         pose_debugging['cluster_obb']
    #     ),
    # ])

    # o3d.visualization.draw_geometries(geometry_list)


def localize_objects(dataset, scene_id, objects, clusters):
    def pose_from_obb(obb):
        T = np.eye(4)
        T[:3, :3] = correct_rotation(obb.R)
        T[:3, -1] = obb.get_center()
        return T

    def rotate_obb(obb, R):
        """Rotate OBB about its local frame"""
        obb_center = obb.get_center()
        obb_R = correct_rotation(obb.R)
        origin = np.zeros((3, 1))

        return copy.deepcopy(aligned_object_obb) \
            .translate(-obb_center, relative=True) \
            .rotate(obb_R.transpose(), origin) \
            .rotate(R, origin) \
            .rotate(obb_R, origin) \
            .translate(obb_center, relative=True)

    object_poses = {}
    for scene_object_name in objects:
        mesh_path = dataset.root / 'models' / scene_object_name / 'visual.ply'
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
        object_obb = o3d_mesh.get_oriented_bounding_box()

        best_fitness = -1
        best_pose = None

        for cluster in clusters:
            cluster_obb = cluster.get_oriented_bounding_box()
            # transform object OBB to cluster OBB
            translation = cluster_obb.get_center() - object_obb.get_center()
            rotation = cluster_obb.R.dot(object_obb.R.transpose())
            aligned_object_obb = copy.deepcopy(object_obb) \
                                    .translate(translation)
            aligned_object_obb = aligned_object_obb.rotate(rotation)

            # rotations to flip the OBB 180
            # since the original rotation might not be correct
            flip_rotations = [
                np.eye(3),
                o3d.geometry.get_rotation_matrix_from_xyz([np.pi, 0, 0]),
                o3d.geometry.get_rotation_matrix_from_yzx([np.pi, 0, 0]),
                o3d.geometry.get_rotation_matrix_from_zyx([np.pi, 0, 0]),
            ]

            object_obbs = [
                copy.deepcopy(rotate_obb(aligned_object_obb, R))
                for R in flip_rotations
            ]

            pose_candidates = [
                pose_from_obb(obb)
                for R, obb in zip(flip_rotations, object_obbs)
            ]

            # normals required for clusters if point-to-plane ICP
            # if not cluster.has_normals():
            #     cluster.estimate_normals(
            #         search_param=o3d.geometry.KDTreeSearchParamHybrid(
            #             radius=0.1, max_nn=30
            #         )
            #     )

            refined_poses = [
                icp_refine_object_poses(
                    dataset, {scene_object_name: pose},
                    cluster, ICP_ALIGNMENT_THRESHOLD, use_normal=False
                )
                for pose in pose_candidates
            ]

            for pose, quality in refined_poses:
                print(quality)
                if quality[scene_object_name]['fitness'] > best_fitness:
                    best_fitness = quality[scene_object_name]['fitness']
                    best_pose = pose[scene_object_name]

        object_poses[scene_object_name] = best_pose

    return object_poses, {
        'object_obb': object_obb, 'cluster_obb': cluster_obb,
        'object_obbs': object_obbs, 'pose_candidates': pose_candidates
    }


def cluster_point_cloud(point_cloud):
    labels = np.array(point_cloud.cluster_dbscan(
        eps=CLUSTER_DENSITY, min_points=CLUSTER_MIN_POINTS, print_progress=True
    ))

    max_label = labels.max()
    print(f"point cloud has {max_label + 1} clusters")

    colors = plt.get_cmap("tab20")(labels / (max_label if max_label > 0 else 1))
    colors[labels < 0] = 0
    colored_pcd = copy.deepcopy(point_cloud)
    colored_pcd.colors = o3d.utility.Vector3dVector(colors[:, :3])

    clusters = []
    for cluster_idx in range(max_label+1):
        clusters.append(point_cloud.select_by_index(
            np.where(labels == cluster_idx)[0].tolist()
        ))
    return clusters, colored_pcd


def remove_objects_from_point_cloud(dataset, point_cloud, object_poses):
    filtered_point_cloud = copy.deepcopy(point_cloud)
    for scene_object_name, T_world_object in object_poses.items():
        if T_world_object is None:
            continue
        mesh_path = dataset.root / 'models' / scene_object_name / \
                        'visual.ply'
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
        o3d_mesh_world = copy.deepcopy(o3d_mesh).transform(T_world_object)
        obb = o3d_mesh_world.get_oriented_bounding_box()
        # inflate the OBB to filter noise around the edges
        obb = obb.scale(1.25, obb.center)
        inds = obb.get_point_indices_within_bounding_box(
            filtered_point_cloud.points
        )
        filtered_point_cloud = filtered_point_cloud.select_by_index(
            inds, invert=True
        )
    return filtered_point_cloud


def get_geometry_list(
    dataset, scene_id, object_poses, camera_poses, point_cloud
):
    geometry_list = []
    for scene_object_name, T_world_object in object_poses.items():
        if T_world_object is None:
            continue
        mesh_path = dataset.root / 'models' / scene_object_name / \
                        'visual.ply'
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
        o3d_mesh_world = copy.deepcopy(o3d_mesh).transform(T_world_object)
        geometry_list.append(
            o3d.geometry.LineSet.create_from_triangle_mesh(o3d_mesh_world)
        )

    geometry_list.append(point_cloud)
    geometry_list.append(
        o3d.geometry.TriangleMesh.create_coordinate_frame(0.1)
    )
    return geometry_list


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/toc_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default=''
    )
    args = parser.parse_args()
    main(**args.__dict__)
