import sys

sys.path.append('Hierarchical-Localization')

import numpy as np
import torch
from tqdm import tqdm
import cv2
import tempfile
from pathlib import Path
import shutil

from hloc import extract_features, match_features
from hloc.extract_features import ImageDataset
from hloc import extractors, matchers
from hloc.utils.base_model import dynamic_load
from hloc.utils.tools import map_tensor


@torch.no_grad()
def get_feature_extractor(conf):
    conf = extract_features.confs[conf]
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    Model = dynamic_load(extractors, conf['model']['name'])
    model = Model(conf['model']).eval().to(device)
    return model


@torch.no_grad()
def get_feature_matcher(conf='superglue'):
    conf = match_features.confs[conf]
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    Model = dynamic_load(matchers, conf['model']['name'])
    model = Model(conf['model']).eval().to(device)
    return model


@torch.no_grad()
def get_images_features(conf, image_dir, model=None, as_half=False):
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    conf = extract_features.confs[conf]
    if model is None:
        Model = dynamic_load(extractors, conf['model']['name'])
        model = Model(conf['model']).eval().to(device)

    loader = ImageDataset(image_dir, conf['preprocessing'])
    loader = torch.utils.data.DataLoader(loader, num_workers=1)

    features_dict = {}

    for data in tqdm(loader):
        pred = model(map_tensor(data, lambda x: x.to(device)))
        pred = {k: v[0].cpu().numpy() for k, v in pred.items()}

        pred['image_size'] = original_size = data['original_size'][0].numpy()
        if 'keypoints' in pred:
            size = np.array(data['image'].shape[-2:][::-1])
            scales = (original_size / size).astype(np.float32)
            pred['keypoints'] = (pred['keypoints'] + .5) * scales[None] - .5

        if as_half:
            for k in pred:
                dt = pred[k].dtype
                if (dt == np.float32) and (dt != np.float16):
                    pred[k] = pred[k].astype(np.float16)

        features_dict[data['name'][0]] = pred

    return features_dict


@torch.no_grad()
def get_image_features(conf, image, model=None, as_half=False):
    """creates temp directory and save image in temp file
    and treats it like a database"""
    tmp_dir = tempfile.mkdtemp()
    tmp_file = tempfile.mkstemp(suffix='.png', dir=tmp_dir)
    cv2.imwrite(tmp_file[1], image)
    features = get_images_features(conf, Path(tmp_dir), model, as_half)
    shutil.rmtree(tmp_dir)
    assert len(features) == 1
    return list(features.values())[0]


@torch.no_grad()
def find_matches(feats0, feats1, model, min_match_score=None):
    device = next(model.parameters()).device
    data = {}
    for k in feats1.keys():
        data[k+'0'] = feats0[k].__array__()
    for k in feats1.keys():
        data[k+'1'] = feats1[k].__array__()
    data = {k: torch.from_numpy(v)[None].float().to(device)
            for k, v in data.items()}

    # some matchers might expect an image but only use its size
    data['image0'] = torch.empty((1, 1,)+tuple(feats0['image_size'])[::-1])
    data['image1'] = torch.empty((1, 1,)+tuple(feats1['image_size'])[::-1])

    pred = model(data)
    matches = pred['matches0'].cpu().numpy().__array__()
    valid = matches > -1
    if min_match_score:
        scores = hfile[pair]['matching_scores0'].__array__()
        valid = valid & (scores > min_match_score)
    matches = np.stack([np.where(valid)[-1], matches[valid]], -1)

    return matches

