#!/usr/bin/env python
import sys
sys.path.append('dataset_toolkit')

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level
from renderer import resize_camera_parameters

import torch
import torch.nn
import torch.nn.functional as F
from torch.utils.data import Dataset, DataLoader

from pytorch3d.io import load_obj, load_objs_as_meshes
from pytorch3d.renderer.cameras import PerspectiveCameras
from pytorch3d.structures import join_meshes_as_batch, Meshes

from pytorch3d.renderer import (
    look_at_view_transform,
    PerspectiveCameras,
    FoVPerspectiveCameras,
    SoftPhongShader,
    SoftSilhouetteShader,
    RasterizationSettings,
    BlendParams,
    MeshRenderer,
    MeshRasterizer,
    PointLights,
    # AmbientLights
)

import pytorch3d.transforms

import argparse
from pathlib import Path
import logging
from shutil import copyfile
import time
import tqdm
import tempfile

import numpy as np
import cv2
import open3d as o3d
import pyrender
import trimesh


# Pytorch3D renderer doesn't work with width != height
# plus reduce image size for memory reasons
NEW_WIDTH = NEW_HEIGHT = 500
NEW_FOCAL_LENGTH_X = 0.85 * NEW_HEIGHT
NEW_FOCAL_LENGTH_Y = 0.85 * NEW_WIDTH
NEW_PRINCIPAL_POINT_X = NEW_WIDTH / 2.0
NEW_PRINCIPAL_POINT_Y = NEW_HEIGHT / 2.0

NUM_EPOCHS = 1
BATCH_SIZE = 2
ITERS_PER_BATCH = 10
DEVICE = 'cuda'
ROTATION_LEARNING_RATE = 2.5e-2
TRANSLATION_LEARNING_RATE = 1e-3
RENDER_FACES_PER_PIXEL = 120
NUM_MESH_TRIANGLES = 30000


def main(dataset_root, scene_name, resize_factor):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    scene_dir = dataset_root / 'scenes' / scene_name

    # visualize_mask_rcnn_mask(dataset, scene_name)

    camera_poses = get_camera_poses(scene_dir)
    if camera_poses is None:
        sys.exit('camera poses not found')

    object_name, object_pose = get_object_pose(scene_dir, camera_poses)

    new_object_pose = refine_object_pose(
        dataset, scene_name, camera_poses,
        object_name, object_pose, resize_factor
    )

    if new_object_pose is not None:
        # backup old object pose
        if (scene_dir / 'object_poses.npy').is_file():
            copyfile(
                scene_dir / 'object_poses.npy',
                scene_dir / f'object_poses_{int(time.time())}.npy'
            )
        np.save(
            scene_dir / 'object_poses.npy', {object_name: new_object_pose}
        )


def refine_object_pose(
        dataset, scene_name, camera_poses, object_name, object_pose, resize_factor
):
    global BATCH_SIZE, DEVICE, NEW_HEIGHT, NEW_WIDTH

    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)

    camera_param = dataset.load_camera_param(scene_id)
    camera_param = resize_camera_parameters(camera_param, resize_factor)

    old_fx, old_fy = camera_param.intrinsic.get_focal_length()
    old_cx, old_cy = camera_param.intrinsic.get_principal_point()
    old_K = K_matrix(old_fx, old_fy, old_cx, old_cy)
    new_K = K_matrix(
        NEW_FOCAL_LENGTH_X, NEW_FOCAL_LENGTH_Y,
        NEW_PRINCIPAL_POINT_X, NEW_PRINCIPAL_POINT_Y,
    )

    rectification_masks = get_rectification_masks(
        old_K, new_K, (NEW_WIDTH, NEW_HEIGHT)
    )
    inv_rectification_masks = get_rectification_masks(
        new_K, old_K,
        (camera_param.intrinsic.width, camera_param.intrinsic.height)
    )

    mask_dataset = MaskDataset(dataset, scene_name, camera_poses, rectification_masks)
    mask_dataloader = DataLoader(
        mask_dataset, batch_size=BATCH_SIZE, shuffle=True,
        num_workers=0, pin_memory=True
    )
    mesh_path = Path(dataset.root) / 'models' / object_name / 'textured.obj'

    return train(mask_dataloader, mesh_path, object_pose)


class Model(torch.nn.Module):
    def __init__(self, mesh, renderer, object_pose):
        global DEVICE

        super().__init__()

        self.mesh = mesh
        self.renderer = renderer

        self.device = torch.device(DEVICE)

        T_world_obj = torch.from_numpy(object_pose) \
                                .float().unsqueeze(0).to(self.device)
        self.register_buffer('T_world_obj', T_world_obj)

        self.trans_increment = torch.nn.Parameter(
            torch.from_numpy(np.array([[0, 0, 0]], dtype=np.float32)) \
                .to(self.device)
        )

        self.so3_log_increment = torch.nn.Parameter(
            torch.from_numpy(np.array([[0, 0, 0]], dtype=np.float32)) \
                .to(self.device)
        )

        # self.so3_log_increment = torch.zeros(
        #     1, 3, dtype=torch.float32, device=self.device, requires_grad=True
        # )

    def forward(self, batch):
        batch_size = batch['camera_pose'].shape[0]

        # same mesh for all camera positions
        meshes = join_meshes_as_batch([self.mesh for _ in range(batch_size)])

        return render_object(
            self.renderer, self.get_updated_pose(),
            batch['camera_pose'], meshes
        )

    def get_updated_pose(self):
        se3_log_increment = torch.cat(
            (self.trans_increment, self.so3_log_increment), dim=-1
        )

        T_increment = se3_exp_map(se3_log_increment)
        T_world_obj_new = self.T_world_obj @ T_increment

        return T_world_obj_new


def reinitialize_model_optimizer(meshes, renderer, object_pose):
    global ROTATION_LEARNING_RATE
    global TRANSLATION_LEARNING_RATE

    model = Model(meshes, renderer, object_pose).to(DEVICE)

    optimizer = torch.optim.SGD(
        [
            {'params': model.trans_increment, 'lr': TRANSLATION_LEARNING_RATE},
            {'params': model.so3_log_increment, 'lr': ROTATION_LEARNING_RATE}
        ],
        lr=0.0, # 1e-2,
        momentum=0.0 # 0.2
    )

    # optimizer = torch.optim.Adam(model.parameters(), lr=5e-3)
    # optimizer = torch.optim.SGD(model.parameters(), lr=5e-3, momentum=0.4)

    return model, optimizer


def train(mask_dataloader, mesh_path, object_pose):
    global BATCH_SIZE, DEVICE, NUM_EPOCHS, NEW_HEIGHT, NEW_WIDTH, NUM_MESH_TRIANGLES

    renderer = get_pytorch3d_renderer(DEVICE, NEW_WIDTH, NEW_HEIGHT)
    pyrender_renderer = get_pyrender_renderer(NEW_WIDTH, NEW_HEIGHT, mesh_path)

    o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
    o3d_mesh = o3d_mesh.simplify_quadric_decimation(NUM_MESH_TRIANGLES)
    print('simplified mesh to {} triangles' \
            .format(np.asarray(o3d_mesh.triangles).shape))

    with tempfile.TemporaryDirectory() as tmp:
        tmp_mesh_path = Path(tmp) / 'tmp.obj'
        o3d.io.write_triangle_mesh(str(tmp_mesh_path), o3d_mesh)

        # Load the obj and ignore the textures and materials.
        verts, faces_idx, _ = load_obj(tmp_mesh_path)

    faces = faces_idx.verts_idx
    mesh = Meshes(verts=[verts.to(DEVICE)], faces=[faces.to(DEVICE)])

    # Load the obj with the textures and materials.
    # mesh = load_objs_as_meshes([mesh_path], device=DEVICE)

    model, optimizer \
            = reinitialize_model_optimizer(mesh, renderer, object_pose)

    cv2.namedWindow('win', cv2.WINDOW_NORMAL | cv2.WINDOW_KEEPRATIO)

    for epoch_idx in range(NUM_EPOCHS):
        for batch_idx, batch in tqdm.tqdm(
                enumerate(mask_dataloader), total=len(mask_dataloader)
        ):
            for iter_idx in range(ITERS_PER_BATCH):
                optimizer.zero_grad()

                batch['mask'] = batch['mask'].to(model.device)
                batch['camera_pose'] = batch['camera_pose'].to(model.device)
                batch['gt_mask'] = batch['mask'].to(model.device)
                batch['depth'] = batch['depth'].to(model.device)

                rendered_images, rendered_depths = model(batch)

                wireframe_images = pyrender_mesh(
                    pyrender_renderer, batch['camera_pose'],
                    model.get_updated_pose()
                )

                gt_masks = batch['mask']
                depths = batch['depth']
                depths = (depths * gt_masks)

                # alpha channel of the RGBA image is the mask
                rendered_masks = rendered_images[..., -1]
                rendered_depths = (rendered_depths * rendered_masks)

                # if object mask too small, don't bother
                if torch.sum((rendered_masks.detach() > 1e-7).float()).item() < 100:
                    print('rendered mask empty')
                    break

                loss = torch.mean(
                    (rendered_masks - gt_masks) ** 2
                    # (rendered_depths - depths) ** 2
                )

                if torch.isnan(loss):
                    print('nan loss')
                    break

                loss.backward()

                if torch.any(torch.isnan(model.trans_increment.grad[0])) \
                        or torch.any(torch.isnan(model.so3_log_increment.grad[0])):
                    print('nan increment')
                    break

                previous_trans_increment = model.trans_increment.clone()
                previous_so3_log_increment = model.so3_log_increment.clone()

                optimizer.step()

                trans_update = torch.norm(
                    (previous_trans_increment - model.trans_increment).view(-1)
                ).item()
                so3_log_update = torch.norm(
                    (previous_so3_log_increment- model.so3_log_increment).view(-1)
                ).item()

                # large update, ignore
                if trans_update > 0.01 or so3_log_update > 0.01:
                    print('large increment')
                    model = None
                    optimizer = None
                    break

                log_train_iter(
                    model, batch, epoch_idx, batch_idx, iter_idx,
                    loss, rendered_images, rendered_depths, wireframe_images
                )

            # rotation optimization is in so3 log map.
            # pose reinitialization is required for large rotation updates
            if model is not None:
                object_pose = model.get_updated_pose().detach().cpu().numpy()[0]
                del optimizer
                del model

            model, optimizer \
                    = reinitialize_model_optimizer(mesh, renderer, object_pose)

    cv2.destroyAllWindows()
    return model.get_updated_pose().detach().cpu().numpy()[0]


def log_train_iter(
        model, batch, epoch_idx, batch_idx, iter_idx,
        loss, rendered_images, rendered_depths, wireframe_images
):
    # if iter_idx != ITERS_PER_BATCH - 1:
    #     return

    se3_log_increment = torch.cat(
        (model.trans_increment, model.so3_log_increment), dim=-1
    )

    # print(
    #     'epoch:', epoch_idx,
    #     'cam:', batch['image_id'],
    #     'loss: ', loss.item(),
    #     'inc:', get_se3_log_magnitude(se3_log_increment),
    #     'grad: ', [
    #         torch.norm(model.trans_increment.grad[0]).item() \
    #            if (model.trans_increment.grad is not None) else None,
    #         torch.norm(model.so3_log_increment.grad[0]).item() \
    #             if (model.so3_log_increment.grad is not None) else None
    #     ]
    # )

    # batch_size = batch['camera_pose'].shape[0]
    # minibatch_idx = iter_idx % batch_size
    minibatch_idx = 0

    rendered_mask = rendered_images.detach().cpu().numpy()[minibatch_idx, ..., -1]
    rgb = batch['rgb'].detach().cpu().numpy()[minibatch_idx]
    gt_mask = batch['mask'].detach().cpu().numpy()[minibatch_idx]

    rendered_mask_colorized = np.zeros_like(rgb)
    rendered_mask_colorized[:, :, 2] = rendered_mask

    wireframe_image = wireframe_images[minibatch_idx]
    viz_img = rgb * np.stack([wireframe_image == 0]*3, axis=-1)

    gt_mask_colorized = np.zeros_like(rgb)
    gt_mask_colorized[:, :, 0] = gt_mask

    # viz_img = cv2.addWeighted(rgb, 1.0, rendered_mask_colorized, 0.4, 0)
    # viz_img = cv2.addWeighted(viz_img, 1.0, gt_mask_colorized, 0.4, 0)

    # viz_img1 = cv2.addWeighted(rgb, 1.0, rendered_mask_colorized, 0.4, 0)
    # viz_img2 = cv2.addWeighted(gt_mask_colorized, 0.75, rendered_mask_colorized, 0.75, 0)
    # viz_img = np.concatenate((viz_img1, viz_img2), axis=1)

    # viz_img = cv2.addWeighted(rendered_mask_colorized, 1.0, gt_mask_colorized, 1.0, 0)
    # viz_img = torch.abs(depths[0] - rendered_depths[0]).detach().cpu().numpy()

    cv2.imshow('win', viz_img)
    cv2.waitKey(1)


class MaskDataset(Dataset):
    def __init__(self, dataset, scene_name, camera_poses, rectification_masks):
        self.dataset = dataset
        self.scene_name = scene_name
        self.camera_poses = camera_poses
        self.rectification_masks = rectification_masks
        self.scene_dir = dataset.root / 'scenes' / scene_name

        scene_name_list = dataset.load_scene_name_list()
        self.scene_id = scene_name_list.index(scene_name)

        self.image_ids = sorted([
            int(img_path.stem.split('_')[-1])
            for img_path in (self.scene_dir / 'mask_rcnn_masks').iterdir()
            if img_path.name.endswith('.png')
        ])
        self.image_ids = [i for i in self.image_ids if i in camera_poses]

    def __len__(self):
        return len(self.image_ids)

    def __getitem__(self, idx):
        image_id = self.image_ids[idx]

        mask = cv2.imread(
            str(self.scene_dir / 'mask_rcnn_masks' / f'color_{image_id}.png'),
            cv2.IMREAD_UNCHANGED
        )
        mask = rectify_image(mask, self.rectification_masks)
        mask = torch.from_numpy(mask.astype(np.float32) / 255.0)

        rgb = self.dataset.load_raw_image(
            scene_id=self.scene_id, image_id=image_id, order = 'BGR'
        )
        rgb = rectify_image(rgb, self.rectification_masks)
        rgb = torch.from_numpy(rgb.astype(np.float32) / 255.0)

        depth_path = self.dataset.root / 'scenes' / self.scene_name \
                            / 'distorted_images' / f'depth_{image_id}.png'
        depth = cv2.imread(str(depth_path), cv2.IMREAD_UNCHANGED)
        depth = rectify_image(depth, self.rectification_masks)
        depth = torch.from_numpy(depth.astype(np.float32) / 1000.0)

        camera_pose = torch.from_numpy(self.camera_poses[image_id]).float()

        return {
            'mask': mask, 'rgb': rgb, 'depth': depth,
            'camera_pose': camera_pose, 'image_id': image_id
        }


def render_object(renderer, T_world_obj, T_world_cameras, meshes):
    global DEVICE, NEW_WIDTH, NEW_HEIGHT
    batch_size = T_world_cameras.shape[0]

    T_world_objs = T_world_obj.expand(batch_size, -1, -1)
    T_obj_cameras = torch.inverse(T_world_objs) @ T_world_cameras

    T_obj_cameras_pytorch3d = change_pose_to_pytorch3d(T_obj_cameras)
    T_cameras_obj_pytorch3d = torch.inverse(T_obj_cameras_pytorch3d)

    # rotation transposed because PyTorch3D's weird convention
    R = T_cameras_obj_pytorch3d[:, :3, :3].permute(0, 2, 1)
    T = T_cameras_obj_pytorch3d[:, :3, -1]

    rendered_images, rendered_depths = renderer(
        meshes, # lights=renderer.lights,
        R=R, T=T
    )

    return rendered_images, rendered_depths


class MeshRendererWithDepth(torch.nn.Module):
    def __init__(self, rasterizer, shader, lights=None):
        super().__init__()
        self.rasterizer = rasterizer
        self.shader = shader
        # self.renderer = MeshRenderer(rasterizer=rasterizer, shader=shader)
        self.lights = lights

    def forward(self, meshes_world, **kwargs) -> torch.Tensor:
        fragments = self.rasterizer(meshes_world, **kwargs)
        images = self.shader(fragments, meshes_world, **kwargs)
        # images = self.renderer(meshes_world, **kwargs)

        return images, fragments.zbuf[..., 0]


def get_pytorch3d_renderer(device, width, height):
    focal_length_tensor = torch.tensor(
        [NEW_FOCAL_LENGTH_X, NEW_FOCAL_LENGTH_Y], device=DEVICE
    ).unsqueeze(0) # .expand(T_world_cameras.shape[0], -1, -1)

    principal_point_tensor = torch.tensor(
        [NEW_PRINCIPAL_POINT_X, NEW_PRINCIPAL_POINT_Y], device=DEVICE
    ).unsqueeze(0) # .expand(T_world_cameras.shape[0], -1, -1)

    # camera always at the origin
    # camera = PerspectiveCameras(
    #     focal_length=focal_length_tensor,
    #     principal_point=principal_point_tensor,
    #     image_size=((NEW_WIDTH, NEW_HEIGHT), ), # (W, H)
    #     # R=R, T=T,
    #     device=DEVICE
    # )

    fov = 2 * np.arctan2(NEW_WIDTH, 2.0 * NEW_FOCAL_LENGTH_X)
    camera = FoVPerspectiveCameras(
        fov=fov, degrees=False, device=DEVICE
    )

    blend_params = BlendParams(sigma=5e-6, gamma=1e-10)
    raster_settings = RasterizationSettings(
        image_size=height,
        # image_size=(height, width),
        blur_radius=np.log(1. / 1e-4 - 1.) * blend_params.sigma,
        # low faces per pixel ok if not using photometric error
        faces_per_pixel=RENDER_FACES_PER_PIXEL,
        # faces_per_pixel=20,
        perspective_correct=False
    )

    # lights = AmbientLights(device=device, ambient_color=[[0.75, 0.75, 0.75]])

    renderer = MeshRendererWithDepth(
        rasterizer=MeshRasterizer(raster_settings=raster_settings, cameras=camera),
        shader=SoftSilhouetteShader(blend_params=blend_params),
        # shader=SoftPhongShader(
        #     device=DEVICE, lights=lights,
        #     blend_params=blend_params, cameras=camera
        # ),
        # lights=lights
    )

    return renderer


def get_pyrender_renderer(width, height, mesh_path):
    cam = pyrender.IntrinsicsCamera(
        fx=NEW_FOCAL_LENGTH_X, fy=NEW_FOCAL_LENGTH_Y,
        cx=NEW_PRINCIPAL_POINT_X, cy=NEW_PRINCIPAL_POINT_Y
    )
    o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
    o3d_mesh = o3d_mesh.simplify_quadric_decimation(1000)
    with tempfile.TemporaryDirectory() as tmp:
        tmp_mesh_path = Path(tmp) / 'tmp.obj'
        o3d.io.write_triangle_mesh(str(tmp_mesh_path), o3d_mesh)

        trimesh_mesh = trimesh.load(tmp_mesh_path)
        pyrender_mesh = pyrender.Mesh.from_trimesh(trimesh_mesh)

    scene = pyrender.Scene(
        ambient_light=np.array([0.35, 0.35, 0.35, 1.0])
    )
    scene.add(pyrender_mesh)

    cam_node = scene.add(cam)
    r = pyrender.OffscreenRenderer(viewport_width=width, viewport_height=height)

    return {'renderer': r, 'scene': scene, 'cam_node': cam_node}


def pyrender_mesh(renderer, T_world_cams, T_world_obj):
    # use OpenGL convention (East-Up-South) instead of East-Down-North
    rotation_correction = np.asarray([
        [1, 0, 0], [0, -1, 0], [0, 0, -1]
    ])
    render_flags = pyrender.constants.RenderFlags.ALL_WIREFRAME | pyrender.constants.RenderFlags.RGBA

    batch_size = T_world_cams.shape[0]
    T_obj_world = torch.inverse(T_world_obj[0])

    images = []
    for batch_idx in range(batch_size):
        T_obj_cam = T_obj_world @ T_world_cams[batch_idx]
        T_obj_cam = T_obj_cam.cpu().detach().numpy()
        T_obj_cam[:3, :3] = T_obj_cam[:3, :3] @ rotation_correction

        renderer['scene'].set_pose(renderer['cam_node'], T_obj_cam)
        rendered_color, rendered_depth = renderer['renderer'].render(
            renderer['scene'], flags=render_flags
        )
        images.append(rendered_depth > 0)

    return images


def get_camera_poses(scene_dir):
    if (scene_dir / 'camera_poses.npy').is_file():
        camera_poses = np.load(
            str(scene_dir / 'camera_poses.npy'), allow_pickle=True
        )[()]
    elif (scene_dir / 'CameraTrajectory.npy').is_file():
        orb_poses = np.loadtxt(scene_dir / 'CameraTrajectory.npy')
        camera_poses = {
            int(pose[0]): xyz_quat_to_mat(*(pose[1:].tolist()))
            for pose in orb_poses
        }
        np.save(scene_dir / 'camera_poses.npy', camera_poses)
    else:
        camera_poses = None

    return camera_poses


def xyz_quat_to_mat(x, y, z, qx, qy, qz, qw):
    mat = Quaternion(qw, qx, qy, qz).transformation_matrix
    mat[:3, -1] = [x, y, z]
    return mat


def get_object_pose(scene_dir, camera_poses):
    if (scene_dir / 'object_poses.npy').is_file():
        object_poses = np.load(
            str(scene_dir / 'object_poses.npy'), allow_pickle=True
        )[()]
        if len(object_poses) > 1:
            sys.exit('multiple objects in the scene!')
        elif len(object_poses) == 0:
            sys.exit('no object in the scene!')

        return next(iter(object_poses.items()))
    else:
        # set object pose directly infront of an arbitrary camera
        T_world_cam = next(iter(camera_poses.values()))
        T_cam_obj = np.eye(4)
        T_cam_obj[2, 3] = 0.4
        T_world_obj = (T_world_cam @ T_cam_obj)

        with open(str(scene_dir / 'object_list.txt'), 'r') as f:
            object_name = f.readline().strip()

        return object_name, T_world_obj


def visualize_mask_rcnn_mask(dataset, scene_name):
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    scene_dir = dataset.root / 'scenes' / scene_name

    image_ids = sorted([
        int(img_path.stem.split('_')[-1])
        for img_path in (scene_dir / 'mask_rcnn_masks').iterdir()
        if img_path.name.endswith('.png')
    ])

    for image_id in image_ids:
        rgb = dataset.load_raw_image(
            scene_id=scene_id, image_id=image_id, order = 'BGR'
        )
        mask = cv2.imread(str(
            scene_dir / 'mask_rcnn_masks' / f'color_{image_id}.png'
        ), cv2.IMREAD_UNCHANGED)
        mask_colorized = np.zeros_like(rgb)
        mask_colorized[:, :, 2] = ((mask != 0) * 255)
        cv2.imshow('win', cv2.addWeighted(rgb, 1.0, mask_colorized, 0.4, 0))
        cv2.waitKey(1)


def get_rectification_masks(old_intrinsics, new_intrinsics, new_size):
    return cv2.initUndistortRectifyMap(
        old_intrinsics, newCameraMatrix=new_intrinsics,
        distCoeffs=None, size=new_size, R=None, m1type=cv2.CV_32FC1
    )


def rectify_image(image, rectification_masks):
    return cv2.remap(
        image, rectification_masks[0], rectification_masks[1], cv2.INTER_LINEAR
    )



def K_matrix(fx, fy, cx, cy):
    return np.asarray([
        [fx, 0, cx],
        [0, fy, cy],
        [0, 0, 1]
    ])


def change_pose_to_pytorch3d(pose):
    '''to PyTorch3D convention (West-Up-North) from East-Down-North'''
    rotation_correction = torch.tensor(
        [[[-1, 0, 0], [0, -1, 0], [0, 0, 1]]],
        device=pose.device, dtype=pose.dtype
    ).expand(pose.shape[0], -1, -1)

    pose = torch.clone(pose)
    pose[:, :3, :3] = pose[:, :3, :3] @ rotation_correction
    return pose


def change_pose_to_opengl(pose):
    '''to OpenGL convention (East-Up-South) from East-Down-North'''
    rotation_correction = np.asarray([
        [1, 0, 0], [0, -1, 0], [0, 0, -1]
    ])
    pose = pose.copy()
    pose[:3, :3] = pose[:3, :3].dot(rotation_correction)
    return pose


def se3_exp_map(se3):
    # transpose cuz pytorch3d's weird
    return pytorch3d.transforms.se3_exp_map(se3).permute(0, 2, 1)


def se3_log_map(SE3):
    # transpose cuz pytorch3d's weird
    return pytorch3d.transforms.se3_log_map(SE3.permute(0, 2, 1))


def get_se3_log_magnitude(se3):
    return [torch.norm(se3[0, :3]).item(), torch.norm(se3[0, 3:]).item()]


def get_pose_diff(pose1, pose2):
    pose_1_2 = torch.inverse(pose1) @ pose2
    return se3_log_map(pose_1_2)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=1)
    args = parser.parse_args()
    main(**args.__dict__)
