import sys
sys.path.append('dataset_toolkit')

from toc_dataset_toolkit import TOC_Dataset

import argparse
import glob
import multiprocessing as mp
import numpy as np
import os
import tempfile
import time
import warnings
import cv2
import tqdm
from pathlib import Path

import torch

from detectron2.config import get_cfg, LazyConfig
from detectron2.engine.defaults import DefaultPredictor, default_setup
from detectron2.data import MetadataCatalog
from detectron2.data.detection_utils import read_image
from detectron2.utils.logger import setup_logger
from detectron2.utils.visualizer import GenericMask, _create_text_labels


def setup_cfg(args):
    # load config from file and command-line arguments
    if args.config_file.endswith('.yaml'):
        cfg = get_cfg()
    elif args.config_file.endswith('.py'):
        # TODO: complete .py configs
        cfg = LazyConfig.load(args.config_file)
        cfg = LazyConfig.apply_overrides(cfg, args.opts)
        default_setup(cfg, args)
    else:
        sys.exit('Invalid config file', args.config_file)

    # To use demo for Panoptic-DeepLab, please uncomment the following two lines.
    # from detectron2.projects.panoptic_deeplab import add_panoptic_deeplab_config  # noqa
    # add_panoptic_deeplab_config(cfg)
    cfg.merge_from_file(args.config_file)
    cfg.merge_from_list(args.opts)
    # Set score_threshold for builtin models
    cfg.MODEL.RETINANET.SCORE_THRESH_TEST = args.confidence_threshold
    cfg.MODEL.ROI_HEADS.SCORE_THRESH_TEST = args.confidence_threshold
    cfg.MODEL.PANOPTIC_FPN.COMBINE.INSTANCES_CONFIDENCE_THRESH = args.confidence_threshold
    cfg.freeze()
    return cfg


def get_parser():
    script_dir = os.path.dirname(os.path.realpath(__file__))
    parser = argparse.ArgumentParser(description="Detectron2 demo for builtin configs")
    parser.add_argument(
        "--config-file",
        default=f"{script_dir}/configs/mask_rcnn_X_152.yaml",
        # default=f"{script_dir}/configs/mask_rcnn_R_101_FPN_400ep_LSJ.py",
        metavar="FILE",
        help="path to config file",
    )
    parser.add_argument(
        "--confidence-threshold",
        type=float,
        default=0.5,
        help="Minimum score for instance predictions to be shown",
    )
    parser.add_argument(
        "--opts",
        help="Modify config options using the command-line 'KEY VALUE' pairs",
        default=[],
        nargs=argparse.REMAINDER,
    )

    parser.add_argument(
        '--dataset-root', type=Path, default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str, default='scene2'
    )
    parser.add_argument(
        '--filter-labels', type=str, nargs='+', default=['bowl']
    )
    return parser


def main(dataset_root, scene_name, predictor, filter_labels):
    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    scene_dir = dataset_root / 'scenes' / scene_name
    image_dir = dataset.root / 'scenes' / scene_name / 'rgb_undistort'
    masks_dir = dataset.root / 'scenes' / scene_name / 'mask_rcnn_masks'
    masks_dir.mkdir(exist_ok=True)

    print('scene', scene_id, scene_name)

    device = torch.device("cpu")
    metadata = MetadataCatalog.get(
        "coco_2017_val", # "__unused", #
    )
    class_names = metadata.get("thing_classes", None)

    for image_path in image_dir.iterdir():
        if not image_path.name.endswith('.png') \
                and not image_path.name.endswith('.jpg'):
            continue
        image_id = int(image_path.stem.split('_')[-1])

        # need to use detectron2's read_image
        rgb = read_image(image_path, format='BGR')
        # rgb = dataset.load_raw_image(
        #     scene_id=scene_id, image_id=image_id, order = 'BGR'
        # )

        predictions = predictor(rgb)
        predictions = predictions["instances"].to(device)

        boxes = predictions.pred_boxes if predictions.has("pred_boxes") else None
        scores = predictions.scores if predictions.has("scores") else None
        classes = predictions.pred_classes.tolist() if predictions.has("pred_classes") else None
        keypoints = predictions.pred_keypoints if predictions.has("pred_keypoints") else None

        boxes = [boxes[i] for i in range(len(boxes))]
        classes = [class_names[i] for i in classes]
        scores = [i.item() for i in scores]

        if predictions.has("pred_masks"):
            masks = np.asarray(predictions.pred_masks)
        else:
            masks = None

        print('found classes', list(zip(classes, scores)))

        selected_predictions = [
            {
                'mask': mask, 'box': box, 'label': label, 'score': score
            }
            for mask, box, label, score in zip(masks, [i for i in boxes], classes, scores)
            if label in filter_labels
        ]

        if len(selected_predictions) >= 1:
            if len(selected_predictions) > 1:
                print('multiple predictions found, choosing best')
            selected_prediction = max(
                selected_predictions,
                # key=lambda x: x['box'].area()
                key=lambda x: x['score']
            )
        else:
            selected_prediction = None
            continue

        cv2.imwrite(
            str(masks_dir / image_path.name),
            selected_prediction['mask'].astype(np.uint8) * 255
        )

        viz_img = rgb.copy()
        for mask in masks:
            mask_colorized = np.zeros_like(viz_img)
            mask_colorized[:, :, 0] = (mask.astype(viz_img.dtype) * 255)
            viz_img = cv2.addWeighted(viz_img, 1.0, mask_colorized, 0.4, 0)
        cv2.imshow('win', viz_img)
        cv2.waitKey(1)


if __name__ == "__main__":
    mp.set_start_method("spawn", force=True)
    args = get_parser().parse_args()
    setup_logger(name="fvcore")
    logger = setup_logger()
    logger.info("Arguments: " + str(args))

    dataset_root = args.dataset_root
    delattr(args, 'dataset_root')

    scene_name = args.scene_name
    delattr(args, 'scene_name')

    filter_labels = args.filter_labels
    delattr(args, 'filter_labels')

    cfg = setup_cfg(args)
    predictor = DefaultPredictor(cfg)

    main(dataset_root, scene_name, predictor, filter_labels)
