#!/usr/bin/env python
import sys
import os

sys.path.append(os.path.dirname(os.path.abspath(__file__)))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'dataset_toolkit'))

import argparse
from copy import deepcopy
from pathlib import Path
import tqdm
import time
from shutil import copyfile

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level
from renderer import resize_camera_parameters
from align_utils import correct_rotation

# import sympy
# import scipy.optimize

import cv2
import numpy as np
import open3d as o3d
from scipy.spatial.transform import Rotation


# number of points to sample from a mesh
SCENE_MESH_SAMPLE_POINTS = 1000000
OBJECT_MESH_SAMPLE_POINTS = 500000

# ICP alignment threshold
ICP_ALIGNMENT_THRESHOLD = 0.05


def main(dataset_root, scene_name):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    scene_object_list = dataset.load_scene_object_list(scene_id)
    object_name = scene_object_list[0]
    scene_dir = dataset_root / 'scenes' / scene_name

    camera_poses = np.load(
        str(scene_dir / 'camera_poses.npy'), allow_pickle=True
    ).item()

    scene_mesh = o3d.io.read_triangle_mesh(
        str(scene_dir / 'fastfusion_reconstruction.ply'),
        enable_post_processing=True
    )

    object_mesh = o3d.io.read_triangle_mesh(
        str(dataset_root / 'models' / object_name / 'textured.obj'),
        enable_post_processing=True
    )

    scene_rotation = get_scene_rotation(scene_mesh, camera_poses)

    # least squares based doesn't give pose in right height
    # object_point = get_best_least_square_point(camera_poses)

    plane_points = filter_plane_points(scene_mesh, camera_poses, scene_rotation)
    object_point, plane_points, plane_height = get_object_point_from_planes(
        plane_points, camera_poses, scene_rotation
    )

    object_pose = np.eye(4)
    object_pose[:3, 3] = object_point.reshape((3,))
    object_pose[:3, :3] = scene_rotation

    # object model's lowest z (not the center) is the object_point
    object_lowest_z = np.asarray(object_mesh.vertices)[:, 2].min()
    object_pose[2, 3] += object_lowest_z

    # object_points = filter_object_points(
    #     scene_mesh, scene_rotation, plane_points, plane_height
    # )

    # if object_points.size:
    #     bounding_box_align_object_points(
    #         object_mesh, object_points, scene_rotation
    #     )

    icp_results = icp_align(object_mesh, scene_mesh, object_pose)
    if icp_results['success']:
        object_pose = icp_results['pose']
    else:
        print('icp failed')

    # backup original pose
    if (scene_dir / 'object_poses.npy').is_file():
        stamp = str(int(time.time()))
        copyfile(
            str(scene_dir / 'object_poses.npy'),
            str(scene_dir / f'object_poses_{stamp}.npy')
        )

    np.save(
        str(scene_dir / 'object_poses.npy'),
        {object_name: object_pose}
    )

    # pcd = o3d.geometry.PointCloud()
    # pcd.points = o3d.utility.Vector3dVector(plane_points)

    # o3d.visualization.draw_geometries([
    #     scene_mesh, pcd,
    #     o3d.geometry.TriangleMesh \
    #             .create_coordinate_frame(size=0.1, origin=[0, 0, 0]) \
    #             .transform(object_pose),
    #     object_mesh.transform(object_pose),
    # ])


def icp_align(
        object_mesh, scene_mesh, initial_pose,
        alignment_threshold=ICP_ALIGNMENT_THRESHOLD, use_normal=True
):
    global OBJECT_MESH_SAMPLE_POINTS, SCENE_MESH_SAMPLE_POINTS

    object_pcd = object_mesh.sample_points_uniformly(
            number_of_points=OBJECT_MESH_SAMPLE_POINTS,
            use_triangle_normal=True
    )
    scene_pcd = scene_mesh.sample_points_uniformly(
            number_of_points=SCENE_MESH_SAMPLE_POINTS,
            use_triangle_normal=True
    )

    try:
        registration = o3d.registration
    except AttributeError:
        registration = o3d.pipelines.registration

    if use_normal and scene_pcd.has_normals():
        # TODO: load color in object model and do colored ICP
        # registration.TransformationEstimationForColoredICP(),
        kwargs = {
            'estimation_method': \
                registration.TransformationEstimationPointToPlane()
        }
    else:
        kwargs = {
            'estimation_method': \
                registration.TransformationEstimationPointToPoint()
        }

    reg_p2l = registration.registration_icp(
        object_pcd, scene_pcd, alignment_threshold, initial_pose,
        criteria=registration.ICPConvergenceCriteria(
            relative_fitness=1e-6,
            relative_rmse=1e-6,
            max_iteration=1000
        ),
        **kwargs
    )

    return {
        'success': is_icp_successful(reg_p2l.transformation, initial_pose),
        'pose': reg_p2l.transformation,
        'fitness': reg_p2l.fitness,
        'inlier_rmse': reg_p2l.inlier_rmse,
        'correspondences': np.asarray(reg_p2l.correspondence_set).shape[0]
    }


def is_icp_successful(icp_pose, initial_pose):
    position_diff =  np.linalg.norm(icp_pose[:3, 3] - initial_pose[:3, 3])
    if position_diff > 1.0:
        return False

    rotation_diff = initial_pose[:3, :3].T @ icp_pose[:3, :3]
    pitch, roll, yaw = Rotation.from_matrix(rotation_diff) \
                            .as_euler('xyz', degrees=True).tolist()

    print('icp update', position_diff, pitch, roll, yaw)
    if np.abs(pitch) > 15 or np.abs(roll) > 15:
        return False

    return True


def filter_object_points(mesh, R_world_enu, plane_points, plane_height):
    global SCENE_MESH_SAMPLE_POINTS

    R_enu_world = R_world_enu.T

    plane_points_enu = (R_enu_world @ plane_points.T).T
    min_x, max_x = np.min(plane_points_enu[:, 0]), np.max(plane_points_enu[:, 0])
    min_y, max_y = np.min(plane_points_enu[:, 1]), np.max(plane_points_enu[:, 1])

    pcd = mesh.sample_points_uniformly(
        number_of_points=SCENE_MESH_SAMPLE_POINTS,
        use_triangle_normal=True
    )
    pcd_enu = deepcopy(pcd).rotate(R_enu_world, center=[0, 0, 0])
    pcd_enu = np.asarray(pcd_enu.points)

    invalid_points = (
        # filter based on table plane height
        ((pcd_enu[:, 2] - plane_height) < 0.0001)
        # filter based of table plane xy bounds
        | (pcd_enu[:, 0] < min_x) | (pcd_enu[:, 0] > max_x)
        | (pcd_enu[:, 1] < min_y) | (pcd_enu[:, 1] > max_y)
    )
    invalid_points = invalid_points.reshape((-1,))
    object_points = np.asarray(pcd.points)[~invalid_points, :]

    return object_points


def bounding_box_align_object_points(object_mesh, object_points, R_world_enu):
    R_enu_world = R_world_enu.T
    object_points_enu = (R_enu_world @ object_points.T).T
    mesh_vertices = np.asarray(object_mesh.vertices)

    enu_2d_orientation = find_2d_bounding_box_orientation(object_points_enu[:, :2])
    enu_2d_extent = find_2d_bounding_box_extent(
        enu_2d_orientation, object_points_enu[:, :2]
    )

    object_2d_orientation = find_2d_bounding_box_orientation(
        mesh_vertices[:, :2]
    )
    object_2d_extent = find_2d_bounding_box_extent(
        object_2d_orientation, mesh_vertices[:, :2]
    )

    if np.sign(object_2d_extent[0] - object_2d_extent[1]) \
            != np.sign(enu_2d_extent[0] - enu_2d_extent[1]):
        # the xy extents don't match, so 90 degree turn to match
        rotation_correction = np.asarray([[0, -1], [1, 0]])
        object_2d_orientation = object_2d_orientation @ rotation_correction

    enu_3d_transform = np.eye(4)
    enu_3d_transform[:2, :2] = enu_2d_orientation
    enu_3d_transform[:3, 3] = object_points_enu.mean(axis=0)

    object_3d_transform = np.eye(4)
    object_3d_transform[:2, :2] = object_2d_orientation
    object_3d_transform[:3, 3] = enu_3d_transform[:3, 3]
    object_3d_transform[0, 3] += 0.1

    pcd_enu = o3d.geometry.PointCloud()
    pcd_enu.points = o3d.utility.Vector3dVector(object_points_enu)

    o3d.visualization.draw_geometries([
        pcd_enu, deepcopy(object_mesh).transform(object_3d_transform),
        o3d.geometry.TriangleMesh \
                .create_coordinate_frame(size=0.1, origin=[0, 0, 0]) \
                .transform(enu_3d_transform),
        o3d.geometry.TriangleMesh \
                .create_coordinate_frame(size=0.1, origin=[0, 0, 0]) \
                .transform(object_3d_transform),
    ])
    exit(0)


def find_2d_bounding_box_extent(R, points):
    points_aligned = (R @ points.T).T
    min_x, max_x = np.min(points_aligned[:, 0]), np.max(points_aligned[:, 0])
    min_y, max_y = np.min(points_aligned[:, 1]), np.max(points_aligned[:, 1])
    return [max_x - min_x, max_y - min_y]


def find_2d_bounding_box_orientation(points):
    """https://stackoverflow.com/a/45276634
    """
    ca = np.cov(points, y=None,rowvar=0,bias=1)
    v, vect = np.linalg.eig(ca)
    tvect = np.transpose(vect)
    return tvect


def get_object_point_from_planes(plane_points, camera_poses, R_world_enu):
    R_enu_world = R_world_enu.T
    plane_points_enu = (R_enu_world @ plane_points.T).T
    camera_points_enu = {
        cam_idx: R_enu_world @ cam_pose[:3, 3].reshape((3, 1))
        for cam_idx, cam_pose in camera_poses.items()
    }

    mean_camera_height = np.mean([
        cam_pose[2, 3] for cam_pose in camera_poses.values()
    ])

    # shift to camera height to get good height range
    shifted_point_heights = plane_points_enu[:, 2] - mean_camera_height

    # find planes at right height from camera
    height_range = np.arange(-4, 0, 0.1)
    height_votes = {i: 0 for i in range(len(height_range))}

    for point_height in shifted_point_heights.tolist():
        height_idx = np.argmin(np.abs(height_range - point_height))
        height_votes[height_idx] += 1

    # filter smaller planes
    height_candidates = [
        height_range[key] for key, value in height_votes.items() if value > 100
    ]

    # pick plane closest to the camera
    height_idx = np.argmin(np.abs(height_candidates))
    best_height = height_candidates[height_idx] + mean_camera_height
    best_points = (np.abs(plane_points_enu[:, 2] - best_height) < 0.1)

    plane_points_enu = plane_points_enu[best_points, :]
    plane_center_enu = np.mean(plane_points_enu, axis=0)
    plane_center = (R_world_enu @ plane_center_enu.reshape((3, 1)))

    plane_points = (R_world_enu @ plane_points_enu.T).T

    return plane_center, plane_points, best_height


def filter_plane_points(mesh, camera_poses, R_world_enu):
    points = np.asarray(mesh.vertices)
    if not mesh.has_vertex_normals():
        mesh.compute_vertex_normals()
    normals = np.asarray(mesh.vertex_normals)

    vertical_vector = R_world_enu[:3, 2].reshape((3, 1))

    normal_cosines = normals @ vertical_vector
    normal_angles = np.rad2deg(np.arccos(normal_cosines)).reshape((-1,))

    valid_points = (normal_angles < 7.5)

    num_images = 200 # 50 #
    total_images = len(camera_poses)
    if (num_images <= 0) or (total_images <= num_images):
        step_size = 1
    else:
        step_size = int(np.floor(float(total_images) / num_images))

    for T_world_cam in tqdm.tqdm(
        list(camera_poses.values())[::step_size],
        desc='filtering object points'
    ):
        T_cam_world = np.linalg.inv(T_world_cam)

        points_cam = (T_cam_world[:3, :3] @ points.T) \
                        + T_cam_world[:3, 3].reshape((3, 1))
        points_cam = points_cam.T
        distances = np.linalg.norm(points_cam, axis=-1)

        # bearing angles
        theta = np.rad2deg(np.arctan2(points_cam[:, 0], points_cam[:, 2]))
        phi = np.rad2deg(np.arctan2(points_cam[:, 1], points_cam[:, 2]))

        invalid_cam_points = (
            (points_cam[:, 2] < 0) # back of camera
                | (np.abs(theta) > 45) | (np.abs(phi) > 45)
                | (distances > 1.5)
        )

        R_enu_cam = np.linalg.inv(R_world_enu) @ T_world_cam[:3, :3]
        points_enu = (R_enu_cam @ points_cam.T).T

        invalid_cam_points = (
            invalid_cam_points
            # height
            | (points_enu[:, 2] < -1.5)
            | (points_enu[:, 2] > 0.0)
        )

        valid_points[invalid_cam_points] = False

    valid_points = points[valid_points, :]
    return valid_points


def get_scene_rotation(scene_mesh, camera_poses):
    scene_obb = scene_mesh.get_oriented_bounding_box()
    scene_rotation = correct_rotation(scene_obb.R)

    flip_rotation = 0
    for T in camera_poses.values():
        # camera Y-axis and scene Z-axis should be close to 180 degrees
        if np.dot(T[:3, 1], scene_rotation[:3, 2]) > 0:
            flip_rotation += 1

    if flip_rotation > 0:
        scene_rotation = scene_rotation \
                @ o3d.geometry.get_rotation_matrix_from_yzx([np.pi, 0, 0])

    return scene_rotation


def get_best_least_square_point(camera_poses):
    x, y, z = sympy.symbols('x, y, z', real=True)
    xyz = sympy.Matrix([x, y, z])
    loss = 0

    num_images = 75
    total_images = len(camera_poses)
    if (num_images <= 0) or (total_images <= num_images):
        step_size = 1
    else:
        step_size = int(np.floor(float(total_images) / num_images))

    for camera_pose in list(camera_poses.values())[::step_size]:
        d = 1.5
        x0, y0, z0 = camera_pose[:3, 3]
        # point along z-axis
        x1, y1, z1 = camera_pose[:3, 3] + (d * camera_pose[:3, 2])
        vec1 = sympy.Matrix([x - x0, y - y0, z - z0])
        vec2 = sympy.Matrix([x - x1, y - y1, z - z1])
        loss += ((vec1.cross(vec2)).norm() / d)**2

    loss_function = sympy.lambdify((x, y, z), loss)

    # jacobian = loss.jacobian(xyz)
    # hessian = sympy.hessian(loss, xyz)

    solution = scipy.optimize.least_squares(
        lambda v: loss_function(v[0], v[1], v[2]), (x1, y1, z1)
    )
    return solution.x


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str,
        default='scene2'
    )
    args = parser.parse_args()
    main(**args.__dict__)
