#!/usr/bin/env python
import sys
sys.path.append('dataset_toolkit')

from hloc_utils import get_images_features, get_image_features, find_matches
from hloc_utils import get_feature_extractor, get_feature_matcher

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

import argparse
import os
from pathlib import Path
import tqdm
import logging
import random

import numpy as np
import sklearn.preprocessing
from scipy.spatial.transform import Rotation

import matplotlib.pyplot as plt
import cv2

import open3d as o3d
import pyrender
import trimesh


def main(
    dataset_root, scene_name, num_rendered_images, num_real_images, resize_factor,
    view_radius_min, view_radius_max, feature_conf,
    min_keypoints_matches, min_images_matches, min_valid_image_percent
):
    set_log_level('WARNING')
    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    feature_extractor = get_feature_extractor(feature_conf)
    feature_matcher = get_feature_matcher()

    print('scene', scene_id, scene_name)

    image_dir = dataset.root / 'scenes' / scene_name / 'rgb_undistort'
    real_images_features = get_images_features(
        feature_conf, image_dir, feature_extractor
    )

    scene_object_list = dataset.load_scene_object_list(scene_id)
    # object_pose_dict = dataset.load_object_pose_dict(scene_id)
    camera_param = dataset.load_camera_param(scene_id)
    camera_param = resize_camera_parameters(camera_param, resize_factor)
    print(
        f'width: {camera_param.intrinsic.width}, '
        f'height: {camera_param.intrinsic.height}, '
        f'f: {camera_param.intrinsic.get_focal_length()}, '
        f'c: {camera_param.intrinsic.get_principal_point()}, '
    )

    all_images_dir = dataset_root / 'scenes' / scene_name_list[scene_id] / 'all_images'
    symlink_images(dataset, scene_name_list[scene_id], all_images_dir, num_real_images)

    T_obj_cams = {}
    for scene_object_name in scene_object_list[-1:]:
        mesh_path = dataset.root / 'models' / scene_object_name / 'visual.ply'
        o3d_mesh = o3d.io.read_triangle_mesh(str(mesh_path))
        o3d_mesh.compute_vertex_normals()
        mesh_vertices = np.asarray(o3d_mesh.vertices)
        geometry_list = [o3d_mesh]

        print('object:', scene_object_name)

        rendered_color_dir = dataset_root / 'scenes' / scene_name_list[scene_id] \
                                / 'rendered' / scene_object_name / 'rgb'
        rendered_color_dir.mkdir(parents=True, exist_ok=True)

        rendered_depth_dir = dataset_root / 'scenes' / scene_name_list[scene_id] \
                                / 'rendered' / scene_object_name / 'depth'
        rendered_depth_dir.mkdir(parents=True, exist_ok=True)

        rendered_pose_dir = dataset_root / 'scenes' / scene_name_list[scene_id] \
                                / 'rendered' / scene_object_name / 'camera_pose'
        rendered_pose_dir.mkdir(parents=True, exist_ok=True)

        # avoid print from hloc (doesn't work, fix it)
        logging.basicConfig(level=logging.WARNING)

        renderer_dict = setup_renderer(dataset, scene_object_name, camera_param)

        view_radius_min, view_radius_max = update_view_radius(
            view_radius_min, view_radius_max, o3d_mesh, camera_param
        )

        # render images
        for image_id in tqdm.tqdm(range(num_rendered_images)):
            is_pose_found = False
            for _ in range(50):
                T = sample_random_pose(view_radius_min, view_radius_max)

                if point_to_cloud_distance(T[:3, -1], mesh_vertices) < 0.01:
                    continue

                geometry_list.append(
                    o3d.geometry.TriangleMesh.create_coordinate_frame(0.1) \
                            .transform(T)
                )

                rendered_color, rendered_depth = render_object(T, renderer_dict)
                rendered_obj_mask = rendered_depth > 0

                # Debug only
                # plt.figure()
                # plt.imshow(rendered_color[..., ::-1])
                # plt.show()

                # validate if the render image good
                valid_percent = find_valid_object_percent(rendered_obj_mask)
                if valid_percent < min_valid_image_percent:
                    continue

                rendered_image_features = get_image_features(
                    feature_conf, rendered_color, feature_extractor
                )
                all_matches = find_all_matches(
                    feature_matcher, real_images_features,
                    rendered_image_features,
                    min_keypoints_matches, min_images_matches
                )

                image_matches = 0
                for image_stem, matches in all_matches.items():
                    if matches.shape[0] >= min_keypoints_matches:
                        image_matches += 1
                        ## debug only
                        # show_matches(
                        #     image_dir, image_stem, matches,
                        #     real_images_features, rendered_image_features,
                        #     rendered_color
                        # )

                if image_matches < min_images_matches:
                    continue

                color_path = rendered_color_dir / (('%.4d' % image_id) + '.png')
                cv2.imwrite(str(color_path), rendered_color)
                symlink_force(
                    color_path,
                    all_images_dir / \
                        (('%s_%.4d' % (scene_object_name, image_id)) + '.png')
                )

                depth_path = rendered_depth_dir / (('%.4d' % image_id) + '.png')
                cv2.imwrite(str(depth_path), rendered_depth)

                pose_path = rendered_pose_dir / (('%.4d' % image_id) + '.npy')
                np.save(str(pose_path), T)
                is_pose_found = True
                break

            if not is_pose_found:
                print('could not find a valid rendering pose')
                exit(1)

        # new renderer will be setup for each object, but not for each image
        renderer_dict['renderer'].delete()

    return scene_name_list[scene_id]
        # # origin frame
        # geometry_list.append(
        #     o3d.geometry.TriangleMesh.create_coordinate_frame(0.2)
        # )
        # # camera frame
        # for T_obj_cam in T_obj_cams[scene_object_name]:
        #     geometry_list.append(
        #         o3d.geometry.TriangleMesh.create_coordinate_frame(0.05) \
        #                 .transform(T_obj_cam)
        #     )
        # o3d.visualization.draw_geometries(geometry_list)
        # print('\n\n')


def resize_camera_parameters(camera_param, resize_factor):
    new_camera_param = o3d.camera.PinholeCameraParameters(camera_param)
    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    new_camera_param.intrinsic.set_intrinsics(
        width=int(camera_param.intrinsic.width/resize_factor),
        height=int(camera_param.intrinsic.height/resize_factor),
        fx=fx/resize_factor,
        fy=fy/resize_factor,
        cx=cx/resize_factor,
        cy=cy/resize_factor
    )
    return new_camera_param


def update_view_radius(view_radius_min, view_radius_max, o3d_mesh, camera_param):
    extent = np.mean(o3d_mesh.get_axis_aligned_bounding_box().get_extent())
    view_radius_min = extent + 0.05
    view_radius_max = (2 * extent) + 0.02

    print('object extent', extent)
    print('view radius', view_radius_min, view_radius_max)

    return view_radius_min, view_radius_max


def point_to_cloud_distance(point, cloud):
    distances = cloud - point.reshape((1, 3))
    distances = np.linalg.norm(distances, axis=1)
    assert(cloud.shape[0] == distances.shape[0])
    min_idx = np.argmin(distances)
    min_distance = distances[min_idx]
    print('point to mesh dist:', min_distance)

    return min_distance


def find_all_matches(
    feature_matcher, real_images_features, rendered_image_features,
    min_keypoints_matches, min_images_matches
):
    matches = {}
    image_matches = 0

    for image_id, real_image_features in list(real_images_features.items())[::10]:
        matches[image_id] = find_matches(
            real_image_features, rendered_image_features, feature_matcher
        )
        if matches[image_id].shape[0] >= min_keypoints_matches:
            image_matches += 1
        if image_matches >= min_images_matches:
            break

    return matches


def show_matches(
    image_dir, image_stem, matches, real_features,
    rendered_features, rendered_color
):
    def to_cv_keypoints(keypoints, matches, order):
        return [
            cv2.KeyPoint(
                keypoints[matches[i, order]][0],
                keypoints[matches[i, order]][1],
                1
            )
            for i in range(matches.shape[0])
        ]

    def to_cv_matches(matches):
        # one-one match since we've filetered out keypoints
        return [
            cv2.DMatch(i, i, 0)
            for i in range(matches.shape[0])
        ]

    real_keypoints = to_cv_keypoints(
        real_features[image_stem]['keypoints'], matches, 0
    )
    rendered_keypoints = to_cv_keypoints(
        rendered_features['keypoints'], matches, 1
    )

    real_image = cv2.imread(str(image_dir / image_stem))
    matched_image = cv2.drawMatches(
        real_image, real_keypoints,
        rendered_color, rendered_keypoints,
        to_cv_matches(matches), real_image
    )
    plt.imshow(matched_image)
    plt.show()


def find_valid_object_percent(rendered_obj_mask):
    ## based on bounding box area
    non_zero_pixs = np.where(rendered_obj_mask)
    if non_zero_pixs[0].size == 0 or non_zero_pixs[1].size == 0:
        return 0.0
    min_x = np.min(non_zero_pixs[0])
    min_y = np.min(non_zero_pixs[1])
    max_x = np.max(non_zero_pixs[0])
    max_y = np.max(non_zero_pixs[1])
    bounding_box_area = (max_x - min_x + 1) * (max_y - min_y + 1)
    total_area = rendered_obj_mask.shape[0] * rendered_obj_mask.shape[1]
    valid_percent = float(bounding_box_area) / total_area * 100

    ## based on number of pixels
    # valid_percent = np.sum(rendered_obj_mask.astype(float)) \
    #                 / rendered_obj_mask.size * 100

    return valid_percent

def symlink_images(dataset, scene_name, new_dir, num_images):
    """Symlinks images to a new dir"""
    scene_name_list = dataset.load_scene_name_list()
    image_dir = dataset.root / 'scenes' / scene_name / 'rgb_undistort'
    new_dir.mkdir(exist_ok=True)

    valid_images = []
    for image_path in image_dir.iterdir():
        if image_path.name.endswith('.png') or image_path.name.endswith('.jpg'):
            valid_images.append(image_path)

    valid_images = sorted(
        valid_images, key=lambda x: int(x.stem.split('_')[-1])
    )

    if num_images > 0:
        valid_images = random.sample(valid_images, num_images)

    for image_path in valid_images:
            symlink_force(image_path, new_dir / image_path.name)


def symlink_force(target, link_name):
    try:
        os.symlink(target, link_name)
    except FileExistsError as e:
        os.remove(link_name)
        os.symlink(target, link_name)


def aggregate_stats(stats):
    agg_stats = {}
    for stat, vals in stats.items():
        agg_stats[stat+'_min'] = np.min(vals)
        agg_stats[stat+'_max'] = np.max(vals)
        agg_stats[stat+'_mean'] = np.mean(vals)
        agg_stats[stat+'_std'] = np.std(vals)

    # for stat, val in agg_stats.items():
    #     print(stat, val)
    return agg_stats


def vector_angle(v1, v2):
    angle = np.arccos(np.dot(v1, v2) / np.linalg.norm(v1) / np.linalg.norm(v2))
    cross = np.cross(v1, v2)
    # viewing vector to decide direction
    vn = np.asarray([0, 0, 1]).transpose()
    if np.dot(vn, cross) < 0:
      angle = -angle

    return angle


def get_pose_noise():
    x = np.random.normal(0, 0.05)
    y = np.random.normal(0, 0.05)
    z = np.random.normal(0, 0.05)
    rx = np.random.normal(0, 10)
    ry = np.random.normal(0, 10)
    rz = np.random.normal(0, 10)

    T_noise = np.eye(4)
    T_noise[:3, -1] = np.asarray([x, y , z]).transpose()
    R_noise = Rotation.from_euler('xyz', [rx, ry, rz], degrees=True) \
                .as_matrix()
    T_noise[:3, :3] = R_noise
    return T_noise


def setup_renderer(dataset, scene_object_name, camera_param):
    scene = pyrender.Scene(
        ambient_light=np.array([1.0, 1.0, 1.0, 1.0])
    )

    mesh_path = dataset.root / 'models' / scene_object_name / \
                    'visual.ply'
    pyrender_mesh = pyrender.Mesh.from_trimesh(trimesh.load(mesh_path))

    # the default material has metallic property and is unnecessarily reflective
    # this removes the reflectiveness
    pyrender_mesh.primitives[0].material.baseColorFactor \
            = np.ones(4).astype(np.float32)

    mesh_node = scene.add(pyrender_mesh)

    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    cam = pyrender.IntrinsicsCamera(fx=fx, fy=fy, cx=cx, cy=cy)

    r = pyrender.OffscreenRenderer(
        viewport_width=camera_param.intrinsic.width,
        viewport_height=camera_param.intrinsic.height
    )

    return {
        'mesh_node': mesh_node, 'cam': cam, 'renderer': r, 'scene': scene
    }


def render_object(cam_pose, renderer_dict):
    cam_pose = cam_pose.copy()
    # use OpenGL convention (East-Up-South) instead of East-Down-North
    rotation_correction = np.asarray([
        [1, 0, 0], [0, -1, 0], [0, 0, -1]
    ])
    cam_pose[:3, :3] = np.dot(cam_pose[:3, :3], rotation_correction)

    scene = renderer_dict['scene']
    r = renderer_dict['renderer']
    cam = renderer_dict['cam']

    cam_node = scene.add(cam, pose=cam_pose)
    rendered_color, rendered_depth = r.render(scene)
    rendered_color = cv2.cvtColor(rendered_color, cv2.COLOR_BGR2RGB)

    # remove the current cam node for the next one to be added
    scene.remove_node(cam_node)

    return rendered_color, rendered_depth


def evaluate_scene_poses(
    dataset, scene_id, scene_object_name,
    scene_object_list, object_pose_dict
):
    """not used anymore since we don't want to use initial object poses"""
    stats = {
        'x': [], 'y': [], 'z': [],
        'rx': [], 'ry': [], 'rz': []
    }
    T_obj_cams = []

    image_dir = dataset.root / 'scenes' / scene_object_name / 'rgb_undistort'

    for image_path in image_dir.iterdir():
        if not image_path.name.endswith('.png') \
                and not image_path.name.endswith('.jpg'):
            continue
        image_id = int(image_path.stem)
        T_world_cam = dataset.load_camera_pose(scene_id, image_id)
        if object_pose_dict[scene_object_name] is None:
            continue

        T_world_obj = object_pose_dict[scene_object_name]
        T_obj_cam = np.dot(np.linalg.inv(T_world_obj), T_world_cam)
        T_obj_cams.append(T_obj_cam)

        R = Rotation.from_matrix(T_obj_cam[:3, :3])
        euler = R.as_euler('xyz', degrees=True)
        stats['x'].append(T_obj_cam[0, 3])
        stats['y'].append(T_obj_cam[1, 3])
        stats['z'].append(T_obj_cam[2, 3])
        stats['rx'].append(euler[0])
        stats['ry'].append(euler[1])
        stats['rz'].append(euler[2])

        # print('object', scene_object_name, 'image', image_id)
        # print(T_obj_cam[:3, -1].transpose())
        # print(euler)
        # print('\n\n')

        # to spherical coords
        # radius = np.linalg.norm(T_obj_cam[:3, -1])
        # x, y, z = (T_obj_cam[:3, -1] / radius).transpose().tolist()
        # azimuth = np.arctan2(y, x)
        # elevation = np.arctan2(np.hypot(x, y), z)
        # look_at_T = look_at_matrix(radius, azimuth, elevation)
        # lookat_frame = o3d.geometry.TriangleMesh \
        #             .create_coordinate_frame(0.1) \
        #             .transform(np.linalg.inv(look_at_T))
        # geometry_list.append(lookat_frame)
        # # print(x, y, z)
        # print(T_obj_cam)
        # print(look_at_T)
        # print('\n\n')

    return T_obj_cams, stats


def sample_random_pose(view_radius_min, view_radius_max):
    view_radius = np.random.uniform(view_radius_min, view_radius_max)
    theta = np.random.uniform(0, np.pi * 2)
    phi = np.random.uniform(0, np.pi / 2)
    x = view_radius * np.sin(phi) * np.cos(theta);
    y = view_radius * np.sin(phi) * np.sin(theta);
    z = view_radius * np.cos(phi)
    pt = np.asarray([x, y, z])
    T = np.eye(4)
    T[:3, :3] = look_at_matrix(pt)
    T[:3, -1] = pt
    return T

def look_at_matrix(pt):
    """adapted from https://github.com/thodan/bop_toolkit/blob/master/bop_toolkit_lib/view_sampler.py"""
    # Azimuth from (0, 2 * pi).
    azimuth = np.arctan2(pt[1], pt[0])
    if azimuth < 0:
      azimuth += 2.0 * np.pi

    # Elevation from (-0.5 * pi, 0.5 * pi).
    a = np.linalg.norm(pt)
    b = np.linalg.norm([pt[0], pt[1], 0])
    elev = np.arccos(b / a)
    if pt[2] < 0:
      elev = -elev

    # Rotation matrix.
    # Adopted from gluLookAt function (uses OpenGL coordinate system):
    # [1] http://stackoverflow.com/questions/5717654/glulookat-explanation
    # [2] https://www.opengl.org/wiki/GluLookAt_code
    f = -np.array(pt)  # Forward direction.
    f /= np.linalg.norm(f)
    u = np.array([0.0, 0.0, 1.0])  # Up direction.
    s = np.cross(f, u)  # Side direction.
    if np.count_nonzero(s) == 0:
      # f and u are parallel, i.e. we are looking along or against Z axis.
      s = np.array([1.0, 0.0, 0.0])
    s /= np.linalg.norm(s)
    u = np.cross(s, f)  # Recompute up.
    R = np.array([[s[0], s[1], s[2]],
                  [u[0], u[1], u[2]],
                  [-f[0], -f[1], -f[2]]])

    # Convert from OpenGL to OpenCV coordinate system.
    R_yz_flip = np.asarray([
        [1, 0, 0],
        [0, -1, 0],
        [0, 0, -1]
    ])
    R = R_yz_flip.dot(R)

    return R.transpose()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument('--scene-name', type=str, default='scene2')
    parser.add_argument('--resize-factor', type=int, default=2)
    parser.add_argument('--view-radius-min', type=float, default=0.1)
    parser.add_argument('--view-radius-max', type=float, default=0.15)
    parser.add_argument('--min-valid-image-percent', type=float, default=3.0)
    parser.add_argument('--num-rendered-images', type=int, default=150)
    parser.add_argument('--num-real-images', type=int, default=-1)
    parser.add_argument('--min-keypoints-matches', type=int, default=15)
    parser.add_argument('--min-images-matches', type=int, default=3)
    parser.add_argument('--feature-conf', type=str, default='superpoint_inloc')
    args = parser.parse_args()
    main(**args.__dict__)
