#!/usr/bin/env python
import sys
import os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'dataset_toolkit'))


from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level
from renderer import resize_camera_parameters
from pose_estimation import render_object_masks

import argparse
from pathlib import Path
import tqdm
import logging
import pickle
import json
import tempfile
from copy import deepcopy
import shutil
import subprocess

import numpy as np
import cv2
import pyrender
import trimesh
import open3d as o3d

import torch
import pytorch3d.io
import pytorch3d.ops
from pytorch3d.structures import Meshes

TRAINING_HEIGHT = 224
TRAINING_WIDTH = 224
TRAINING_K = [
    248.0, 0, 112,
    0, 248.0, 112,
    0, 0, 1
]

# distance between camera and object
TRAINING_MIN_DISTANCE = 0.605
TRAINING_MAX_DISTANCE = 2.5


def main(dataset_root, meshmvs_dataset_root, resize_factor, binvox_exec):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()

    meshmvs_dataset_root.mkdir(exist_ok=True, parents=True)

    depth_min = np.inf
    depth_max = np.NINF

    for scene_name in sorted(deepcopy(scene_name_list)):
        print('processing', scene_name)

        scene_id = scene_name_list.index(scene_name)
        scene_dir = dataset_root / 'scenes' / scene_name

        meshmvs_scene_dir = meshmvs_dataset_root / scene_name
        meshmvs_scene_dir.mkdir(exist_ok=True)

        scene_params = load_scene_params(scene_dir)
        if scene_params is None:
            continue

        save_gt_labels(
            dataset_root, scene_name, scene_dir, meshmvs_scene_dir,
            scene_params['camera_poses'], scene_params['object_poses']
        )

        K_new = np.asarray(TRAINING_K).reshape((3, 3))
        # print('K_new', '\n', K_new)
        np.save(str(meshmvs_scene_dir / 'meshmvs_camK.npy'), K_new)

        scene_depth_min, scene_depth_max = save_training_images(
            dataset, scene_name, scene_id,
            scene_dir, meshmvs_scene_dir, resize_factor,
            scene_params['K'], scene_params['distortion_coeff'], K_new
        )
        depth_min = min(depth_min, scene_depth_min)
        depth_max = max(depth_max, scene_depth_max)
        if not (meshmvs_scene_dir / 'meshmvs_camera_poses.npy').is_file():
            continue

        save_voxels(meshmvs_scene_dir, binvox_exec)

    print('depth min/max:', depth_min, depth_max)


def load_scene_params(scene_dir):
    if (scene_dir / 'camera_poses.npy').is_file():
        camera_poses = np.load(
            str(scene_dir / 'camera_poses.npy'), allow_pickle=True
        )[()]
    else:
        None

    if (scene_dir / 'object_poses.npy').is_file():
        object_poses = np.load(
            str(scene_dir / 'object_poses.npy'), allow_pickle=True
        )[()]
    else:
        None

    intrinsics_path = scene_dir / 'color_camK.npy'
    distortion_coeff_path = scene_dir / 'color_cam_distortion_coeffs.npy'

    if (not intrinsics_path.is_file()) or (not distortion_coeff_path.is_file()):
        return None

    K = np.load(intrinsics_path).reshape((3, 3))
    distortion_coeff = np.load(distortion_coeff_path).reshape((-1,))
    assert(K.size == 9)

    return {
        'camera_poses': camera_poses,
        'object_poses': object_poses,
        'distortion_coeff': distortion_coeff, 'K': K,
    }


def save_training_images(
        dataset, scene_name, scene_id,
        scene_dir, meshmvs_scene_dir, resize_factor,
        K, distortion_coeff, K_new
):
    camera_poses = np.load(
        str(meshmvs_scene_dir / 'meshmvs_camera_poses.npy'), allow_pickle=True
    )[()]

    object_poses = np.load(
        str(meshmvs_scene_dir / 'meshmvs_object_poses.npy'), allow_pickle=True
    )[()]

    training_images_dir = meshmvs_scene_dir / 'meshmvs_training_images'
    training_images_dir.mkdir(exist_ok=True)

    # TODO: find if we need to undistort depth
    map1_depth, map2_depth = cv2.initUndistortRectifyMap(
        K, np.zeros((4,)), np.eye(3), K_new,
        (TRAINING_WIDTH, TRAINING_HEIGHT), cv2.CV_32FC1
    )

    map1_rgb, map2_rgb = cv2.initUndistortRectifyMap(
        K, distortion_coeff, np.eye(3), K_new,
        (TRAINING_WIDTH, TRAINING_HEIGHT), cv2.CV_32FC1
    )

    num_invalid_depths = 0
    depth_min = np.inf
    depth_max = np.NINF

    rendered_depths = render_depths(
        dataset, scene_id, meshmvs_scene_dir / 'meshmvs_model.obj',
        camera_poses, object_poses, resize_factor
    )

    for img_id, rendered_depth in tqdm.tqdm(
        rendered_depths.items(), desc='saving training images'
    ):
        # depth_filename = scene_dir / 'distorted_images' / f'depth_{img_id}.png'
        # if not depth_filename.is_file():
        #     continue

        # depth_mm = cv2.imread(str(depth_filename), cv2.IMREAD_UNCHANGED)
        # if depth_mm is None:
        #     continue

        # depth_m = depth_mm.astype(np.float32) / 1000.0

        # valid_depths = (np.abs(rendered_depth - depth_m) < 0.5)
        # mask = (rendered_depth > 0) & valid_depths

        # invalid_depths = (rendered_depth > 0) & (~valid_depths)
        # num_invalid_depths += invalid_depths.astype(float).sum()

        # object_depth_mm = (depth_mm * mask).astype(depth_mm.dtype)

        # USES RENDERED DEPTH, cuz real depth noisy and sparse
        depth_m = rendered_depth
        object_depth_mm = (rendered_depth * 1000).astype(np.uint16)
        mask = (rendered_depth > 0)

        object_depth_mm= cv2.remap(
            object_depth_mm, map1_depth, map2_depth,
            interpolation=cv2.INTER_NEAREST
        )

        # all depth values zero, ignore
        if np.count_nonzero(object_depth_mm) < 10:
            continue

        cv2.imwrite(
            str(training_images_dir / f'depth_{img_id}.png'), object_depth_mm
        )

        object_depth_m = (depth_m * mask).astype(np.float32)
        # zero depths to NaN to avoid in min/max consideration
        object_depth_m[object_depth_m == 0] = np.NAN
        # find min/max depth
        depth_min = min(depth_min, np.nanmin(object_depth_m))
        depth_max = max(depth_max, np.nanmax(object_depth_m))

    for img_id in tqdm.tqdm(camera_poses.keys()):
        rgb_filename = scene_dir / 'distorted_images' / f'color_{img_id}.png'
        rgb = cv2.imread(str(rgb_filename))
        if rgb is None:
            continue

        object_depth_mm = cv2.imread(
            str(training_images_dir / f'depth_{img_id}.png') ,
            cv2.IMREAD_UNCHANGED
        )
        if object_depth_mm is None:
            continue
        mask = (object_depth_mm > 0)

        rgb_undistorted = cv2.undistort(rgb, K, distortion_coeff) # , K_prop)
        rgb_undistorted = cv2.remap(
            rgb_undistorted, map1_rgb, map2_rgb,
            interpolation=cv2.INTER_NEAREST
        )

        object_rgb = (rgb_undistorted * mask[..., np.newaxis]).astype(rgb.dtype)
        object_rgba = cv2.cvtColor(object_rgb, cv2.COLOR_BGR2BGRA)
        object_rgba[:, :, 3] = (mask.astype(rgb.dtype) * 255)

        object_rgba = cv2.resize(
            object_rgba, (TRAINING_WIDTH, TRAINING_HEIGHT)
        )

        cv2.imwrite(
            str(training_images_dir / f'color_{img_id}.png'), object_rgba
        )

    print(
        scene_name, 'invalid depths/image',
        float(num_invalid_depths) / len(camera_poses)
    )

    print(scene_name, 'depth min/max:', depth_min, depth_max)
    return depth_min, depth_max


def render_depths(
        dataset, scene_id, mesh_path, camera_poses, object_poses, resize_factor
):
    rendered_depths = {}

    # use OpenGL convention (East-Up-South) instead of East-Down-North
    rotation_correction = np.asarray([
        [1, 0, 0], [0, -1, 0], [0, 0, -1]
    ])

    try:
        camera_param = dataset.load_camera_param(scene_id)
    except FileNotFoundError as e:
        print(f'error reading camera params of {mesh_path}', e)
        return rendered_depths
    camera_param = resize_camera_parameters(camera_param, resize_factor)
    print(
        f'width: {camera_param.intrinsic.width}, '
        f'height: {camera_param.intrinsic.height}, '
        f'f: {camera_param.intrinsic.get_focal_length()}, '
        f'c: {camera_param.intrinsic.get_principal_point()}, '
    )

    fx, fy = camera_param.intrinsic.get_focal_length()
    cx, cy = camera_param.intrinsic.get_principal_point()
    cam = pyrender.IntrinsicsCamera(fx=fx, fy=fy, cx=cx, cy=cy)

    pyrender_meshes = {}
    for scene_object_name, T_world_object  in object_poses.items():
        trimesh_mesh = trimesh.load(mesh_path)
        trimesh_mesh.apply_transform(T_world_object)
        pyrender_meshes[scene_object_name] \
            = pyrender.Mesh.from_trimesh(trimesh_mesh)

    logging.info('rendering segmentation masks')

    scene = pyrender.Scene(
        ambient_light=np.array([0.35, 0.35, 0.35, 1.0])
    )

    for scene_object_name, T_world_object in object_poses.items():
        if T_world_object is None:
            continue
        scene.add(pyrender_meshes[scene_object_name])

    cam_node = scene.add(cam)
    r = pyrender.OffscreenRenderer(
        viewport_width=camera_param.intrinsic.width,
        viewport_height=camera_param.intrinsic.height
    )

    for image_id, T_world_cam in tqdm.tqdm(camera_poses.items()):
        geometry_list = []

        cam_pose = T_world_cam.copy()
        cam_pose[:3, :3] = np.dot(cam_pose[:3, :3], rotation_correction)
        scene.set_pose(cam_node, cam_pose)

        rendered_color, rendered_depth = r.render(scene)
        # rendered_color = cv2.cvtColor(rendered_color, cv2.COLOR_BGR2RGB)
        # seg_mask = rendered_depth > 0

        rendered_depths[image_id] = rendered_depth

    r.delete()

    return rendered_depths


@torch.no_grad()
def save_gt_labels(
        dataset_root, scene_name, scene_dir, meshmvs_scene_dir,
        camera_poses, object_poses
    ):
    assert(len(object_poses) == 1)
    object_name, object_pose = next(iter(object_poses.items()))

    model_dir = dataset_root / 'models' / object_name
    model_file = model_dir / 'visual.ply'
    if not model_file.is_file():
        print(f'{model_file} not found')
        return
    print(model_file)

    o3d_mesh = o3d.io.read_triangle_mesh(str(model_file))
    scale = get_best_scale(o3d_mesh, camera_poses, object_pose)

    o3d_mesh = o3d_mesh.scale(scale, center=[0, 0, 0])
    new_mesh_path = meshmvs_scene_dir / 'meshmvs_model.obj'
    o3d.io.write_triangle_mesh(str(new_mesh_path), o3d_mesh)

    try:
        gt_mesh = pytorch3d.io.load_objs_as_meshes(
            [new_mesh_path], load_textures=False
        )
    except UnicodeDecodeError as e:
        print(e)
        return

    points, normals = pytorch3d.ops.sample_points_from_meshes(
        gt_mesh, num_samples=9000, return_normals=True
    )
    gt_labels = torch.cat((points, normals), dim=-1) \
                        .squeeze(0).cpu().detach().numpy()

    with open(meshmvs_scene_dir / 'meshmvs_gt_labels.dat', 'wb') as f:
        pickle.dump(gt_labels, f)

    with open(meshmvs_scene_dir / 'meshmvs_info.json', 'w') as f:
        json.dump({'scale': scale}, f)

    meshmvs_object_pose = object_pose.copy()
    meshmvs_object_pose[:3, -1] *= scale
    np.save(
        str(meshmvs_scene_dir / 'meshmvs_object_poses.npy'),
        {object_name : meshmvs_object_pose}
    )

    meshmvs_camera_poses = {}
    for image_id, pose in camera_poses.items():
        new_pose = pose.copy()
        new_pose[:3, -1] *= scale
        meshmvs_camera_poses[image_id] = new_pose

    np.save(
        str(meshmvs_scene_dir / 'meshmvs_camera_poses.npy'),
        meshmvs_camera_poses
    )


def get_best_scale(o3d_mesh, camera_poses, T_world_obj):
    z_min = np.inf
    z_max = np.NINF
    z_spans = []

    # simplify for faster processing
    o3d_mesh = o3d_mesh.simplify_quadric_decimation(15000)

    for T_world_cam in camera_poses.values():
        T_cam_world = np.linalg.inv(T_world_cam)
        T_cam_obj = T_cam_world @ T_world_obj
        o3d_mesh_cam = deepcopy(o3d_mesh).transform(T_cam_obj)
        coords = np.asarray(o3d_mesh_cam.vertices)
        near = np.min(coords[:, 2])
        far = np.max(coords[:, 2])
        z_min = min(z_min, near)
        z_max = max(z_max, far)
        z_spans.append(far - near)

    # fit the z-span of the object
    # z_span_avg = np.mean(z_spans)
    # best_z_span = (TRAINING_MIN_DISTANCE + TRAINING_MAX_DISTANCE) / 2.0
    # scale = best_z_span / z_span_avg

    # fit the min distance of the object
    scale = TRAINING_MIN_DISTANCE / z_min

    # print(z_min * scale, z_max * scale)
    # print((z_max - z_min) * scale)
    return scale


def save_voxels(scene_dir, binvox_exec):
    voxels_dir = scene_dir / 'meshmvs_voxels'
    voxels_dir.mkdir(exist_ok=True)

    if (voxels_dir / 'voxels_world.binvox').is_file():
        return

    model_file = scene_dir / 'meshmvs_model.obj'

    # o3d_mesh = o3d.io.read_triangle_mesh(str(model_file))
    # print(o3d_mesh.get_axis_aligned_bounding_box())
    # o3d.visualization.draw_geometries([
    #     o3d_mesh,
    #     o3d.geometry.TriangleMesh.create_coordinate_frame(size=0.1, origin=[0, 0, 0])
    # ])
    # exit(0)

    ret = run_cmd([
        binvox_exec, '-d', '128', '-t', 'binvox', 'ri', '-dc',
        # '-bb', '-0.1', '-0.1', '-0.1', '0.1', '0.1', '0.1',
        str(model_file)
    ])
    if ret == 0:
        shutil.move(
            scene_dir / 'meshmvs_model.binvox',
            voxels_dir / f'voxels_world.binvox'
        )


def run_cmd(cmd):
    ret = subprocess.call(cmd)
    if ret != 0:
        print('Problem with ', ' '.join(cmd))
    return ret


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument(
        '--meshmvs-dataset-root', type=Path,
        default='/root/3d_dataset'
    )
    parser.add_argument('--resize-factor', type=int, default=1)
    parser.add_argument('--binvox-exec', type=str, default='binvox')
    args = parser.parse_args()
    main(**args.__dict__)
