#!/usr/bin/env python
import sys
import os
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..'))
sys.path.append(os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'dataset_toolkit'))

from toc_dataset_toolkit import TOC_Dataset
from toc_dataset_toolkit.utils.logging import set_log_level

import argparse
from copy import deepcopy
import itertools
from pathlib import Path
import tqdm

import cv2
import numpy as np
import open3d as o3d
from pyquaternion import Quaternion

from pose_estimation import (
    parse_colmap_poses, parse_colmap_points, parse_colmap_keypoints,
    get_rotation_error
)


def main(dataset_root, scene_name, resize_factor):
    set_log_level('WARNING')

    dataset = TOC_Dataset(root=dataset_root)
    scene_name_list = dataset.load_scene_name_list()
    scene_id = scene_name_list.index(scene_name)
    print('scene', scene_id, scene_name)

    scene_dir = dataset_root / 'scenes' / scene_name
    colmap_output = scene_dir / 'colmap_output_all_images_2'
    pose_filename = colmap_output / 'images.txt'
    points_filename = colmap_output / 'points3D.txt'

    camera_poses, rendered_poses = parse_colmap_poses(pose_filename)
    point_cloud = parse_colmap_points(points_filename)
    camera_keypoints, rendered_keypoints = parse_colmap_keypoints(pose_filename)

    # only one object
    object_name = next(iter(rendered_poses.keys()))
    rendered_poses = next(iter(rendered_poses.values()))
    rendered_keypoints = next(iter(rendered_keypoints.values()))

    for camera_id, rendered_id in tqdm.tqdm(itertools.product(
        camera_poses.keys(), rendered_poses.keys()
    )):
        camera_pose = camera_poses[camera_id]
        rendered_pose = rendered_poses[rendered_id]

        translation_diff = np.linalg.norm(camera_pose[:3, -1] - rendered_pose[:3, -1])
        rotation_diff = get_rotation_error(camera_pose[:3, :3], rendered_pose[:3, :3])

        if translation_diff < 3.0 and np.all(np.abs(rotation_diff) < 30):
            camera_image_filename \
                = scene_dir / 'rgb_undistort' / f'color_{camera_id}.png'
            camera_image = cv2.imread(str(camera_image_filename))
            camera_kpts_dict = camera_keypoints[camera_id]
            camera_pt3d_ids = set(camera_kpts_dict.keys())

            rendered_image_filename = (
                scene_dir / 'rendered' / object_name /
                'rgb' / f'{rendered_id:04d}.png'
            )
            rendered_image = cv2.imread(str(rendered_image_filename))
            rendered_kpts_dict = rendered_keypoints[rendered_id]
            rendered_pt3d_ids = set(rendered_kpts_dict.keys())

            camera_kpts = []
            rendered_kpts = []
            for pt3d_id in list(camera_pt3d_ids.intersection(rendered_pt3d_ids))[:15]:
                if pt3d_id == -1:
                    continue
                camera_kpts.append(cv2.KeyPoint(*camera_kpts_dict[pt3d_id], 1))
                rendered_kpts.append(cv2.KeyPoint(*rendered_kpts_dict[pt3d_id], 1))

            if len(camera_kpts) < 10:
                continue
            out_image = deepcopy(camera_image)
            out_image = cv2.drawMatches(
                camera_image, camera_kpts, rendered_image, rendered_kpts,
                [cv2.DMatch(i, i, 1) for i in range(len(camera_kpts))],
                out_image
            )
            cv2.imshow('win', out_image)
            cv2.waitKey()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--dataset-root', type=Path, default='/root/3d_dataset'
    )
    parser.add_argument(
        '--scene-name', type=str, default='scene2'
    )
    parser.add_argument('--resize-factor', type=int, default=2)
    args = parser.parse_args()
    main(**args.__dict__)
